"""
Author: David Cronkite, GHRI
Date: 20dec12
Purpose:
    Methods for simplying the creation of a logfile
"""

import logging
from logging.handlers import RotatingFileHandler
import os


def setup(name=__name__, logdir='log', console=True, text=True, loglevel='DEBUG', logfile=None):
    """
    function:
        prepares a logger that can print to both the console and a logfile
    parameters:
        name - the logger's name (for the call logging.getLogger(name))
        logdir - location for the output logfile (default is current dir)
        console - print to console
        text - print to logfile
        loglevel - level for logger (logging.DEBUG)
        logfile - name of logfile (default=name + '.log')
    use:
        logging.config.dictConfig( mylogger.setup(name, etc.) )
        :param console:
        :param text:
        :param loglevel:
        :param logfile:
        :param logdir:
        :param name:
    """
    logdir = os.path.abspath(logdir)
    os.makedirs(logdir, exist_ok=True)
    if not logfile:
        logfile = name + '.log'

    handler_list = []
    if console:
        handler_list.append('console')
    if text:
        handler_list.append('text')

    config_dict = {'version': 1,
                   'disable_existing_loggers': False,
                   'formatters': {
                       'standard': {
                           'format': '%(asctime)s - %(levelname)s: %(message)s',
                       },
                   },
                   'handlers': {
                       'text': {
                           'level': loglevel,
                           'class': 'logging.handlers.RotatingFileHandler',
                           'backupCount': 5,
                           'filename': os.path.join(logdir, logfile),
                           'formatter': 'standard'
                       },
                       'console': {
                           'class': 'logging.StreamHandler',
                           'level': loglevel,
                           'formatter': 'standard',
                           'stream': 'ext://sys.stdout'
                       }
                   },
                   'loggers': {
                       name: {
                           'level': loglevel,
                           'handlers': handler_list
                       }
                   },
                   'root': {
                       'level': loglevel,
                       'handlers': handler_list}

                   }

    return config_dict


def setup_local(name=__name__, logdir='log', console=True, text=True, loglevel=logging.DEBUG, logfile=None):
    """
        function:
                prepares a logger that can print to both the console and a logfile
        parameters:
                name - the logger's name (for the call logging.getLogger(name))
                logdir - location for the output logfile (default is current dir)
                console - print to console
                text - print to logfile
                loglevel - level for logger (logging.DEBUG)
                logfile - name of logfile (default=name + '.log')
        use:
                mylogger.setup(name, etc.)
                logger = mylogger.getLogger(name)
                :param name:
                :param logdir:
                :param console:
                :param text:
                :param loglevel:
                :param logfile:
        """
    logdir = os.path.abspath(logdir)
    os.makedirs(logdir, exist_ok=True)
    logger = logging.getLogger(name)
    logger.setLevel(loglevel)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s: %(message)s')

    if not logfile:
        logfile = name + '.log'

    if text:
        txt_handler = RotatingFileHandler(os.path.join(logdir, logfile), backupCount=5)
        txt_handler.doRollover()
        txt_handler.setFormatter(formatter)
        logger.addHandler(txt_handler)

    if console:
        scrn_handler = logging.StreamHandler()
        scrn_handler.setFormatter(formatter)
        logger.addHandler(scrn_handler)

    return logging.getLogger(name)


def get_logger(name=__name__):
    return logging.getLogger(name)


def resolve_verbosity(verbosity_level):
    """
    Resolves verbosity based on the assumption that:
    3=Debug
    2=info
    1=warning
    0=error

    Returns logging.[LOGLEVEL]
    :param verbosity_level:
    """
    if verbosity_level <= 0:
        loglevel = logging.ERROR
    elif verbosity_level == 1:
        loglevel = logging.WARNING
    elif verbosity_level == 2:
        loglevel = logging.INFO
    else:
        loglevel = logging.DEBUG
    return loglevel



from .base import Dictionary, Document, Output
from .csv import CsvDictionary, CsvDocument, CsvOutput
from .jsonl import JsonlOutput
from .sas import SasDictionary, SasDocument
from .sql import SqlDictionary, SqlDocument, SqlOutput



"""
Interact with database through SQL Alchemy library
"""
import argparse
import sqlalchemy as sqla
from sqlalchemy import MetaData
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

DRIVER_TO_DBAPI = {
    'SQL Server': 'mssql+pyodbc',
    'MySQL': 'mysql+pyodbc',  # untested, subject to change
    'PostgreSQL': 'postgresql+psycopg2',  # untested, subject to change
}

Base = declarative_base()

CURRENT_ENGINE = {
}


class TableWrapper(object):
    """Wrap this around tables obtained from automap base to get easier column access"""
    def __init__(self, table):
        self.table = table

    def __getitem__(self, item):
        return self.table.__dict__[item]  # but this includes some other items
        # return self.table.__table__._columns[item]  # relies on private access

    def __getattr__(self, item):
        return getattr(self.table, item)


def get_new_session(engine=None, name=None):
    if engine:
        return sessionmaker(bind=engine)()
    else:
        return sessionmaker(bind=CURRENT_ENGINE[name])()


def add_dbargs_to_argparser(parser=None):
    """

    :param parser: existing argparser, or None to create one
    :return: argparse.ArgumentParser
    """
    if not parser:
        parser = argparse.ArgumentParser(fromfile_prefix_chars='@!')
    parser.add_argument('--dbapi',
                        help='DBAPI (see SqlAlchemy doco).')
    parser.add_argument('--database', required=True,
                        help='Database name.')
    parser.add_argument('--server', required=True,
                        help='Server name to connect to.')
    parser.add_argument('--driver', required=True,
                        help='Driver. No need to specify dbapi if driver in list: {}'.format(
                            ', '.join(DRIVER_TO_DBAPI.keys())))
    return parser


def get_engine_from_args(argstring, name=None):
    """

    :param argstring: sys.argv, or equivalent
    :param name:
    :return:
    """
    return get_engine(name=name,
                      **vars(add_dbargs_to_argparser().parse_known_args(argstring)[0]))


def get_tables(engine, *tablenames, schema='dbo'):
    metadata = MetaData(engine, schema=schema)
    autobase = automap_base(metadata=metadata)
    autobase.prepare(engine, reflect=True)
    tables = []
    for tablename in tablenames:
        if tablename in autobase.classes.keys():
            try:
                tables.append(autobase.classes[tablename])
                continue  # don't throw error if reached
            except AttributeError as ae:
                print(ae)
        # throw error if table name found
        raise ValueError('Table {}.{} does not exist or does not have a primary key.'
                         ' Add a primary key (ALTER TABLE {} ADD ID INTEGER IDENTITY(1,1)'
                         ' Or set the primary key '
                         ' ALTER TABLE {} ALTER COLUMN ID INTEGER NOT NULL '
                         ' ALTER TABLE {} ADD CONSTRAINT {}_pk PRIMARY KEY (ID)'
                         ''.format(schema, tablename, tablename, tablename, tablename, tablename))
    return tables


def get_engine(name=None, **kwargs):
    """

    :param name: name to give new connection; name of connection to retrieve
    :param kwargs:
        * dbapi
        * database
        * server
        * driver
        * pool_size
    :return:
    """
    if CURRENT_ENGINE and 'server' not in kwargs:
        try:
            return CURRENT_ENGINE[name]
        except KeyError:
            pass
        if len(CURRENT_ENGINE) == 1:
            return CURRENT_ENGINE[list(CURRENT_ENGINE.keys())[0]]
        else:
            raise ValueError('Multiple connections: unknown which connection to retrieve.')
    else:
        load_session(name=name, **kwargs)
        return CURRENT_ENGINE[name]


def load_session(pool_size=5, name=None, connection_string=None, **kwargs):
    """
    Create a session, with the specified connection.
    :param: connection string or tuple(driver, server, database)
    :return:
    """
    global CURRENT_ENGINE
    connection_string = connection_string or get_connection_string(**kwargs)
    if pool_size == 0:
        CURRENT_ENGINE[name] = sqla.create_engine(connection_string, poolclass=NullPool)
    else:
        CURRENT_ENGINE[name] = sqla.create_engine(connection_string, pool_size=pool_size)
    return sessionmaker(bind=CURRENT_ENGINE[name])()


def get_connection_string(**kwargs):
    try:
        return kwargs['connection_string']
    except KeyError:
        return get_connection_string_from_args(**kwargs)


def get_connection_string_from_args(dbapi=None, database=None, server=None, driver=None, **kwargs):
    if not database or not server or not driver:
        raise ValueError('Database configuration requires driver, server, and database to be specified.')

    if driver and not dbapi:
        try:
            dbapi = DRIVER_TO_DBAPI[driver]
        except KeyError:
            raise ValueError('Could not infer DBAPI from driver "{}".'.format(driver))

    return '{dbapi}://@{server}/{database}?driver={driver}'.format(
        dbapi=dbapi, server=server, database=database, driver=driver)



"""
David Cronkite

Purpose:
    Automatically create a minimal set of RegExs from a sequence of terms.
    The identity of the terms is retained.
"""
# regex should replace/expand re
# may require a new flag to get
# default regex behavior after this
# transformation
import regex as re


def get_affix_morphs():
    """
    list of tuple(regex, replacement string, count) for modifying 
        the front of a term
    count represents the number of letters which are becoming
        optional (not _new_ optional ones)
    """
    return [(re.compile(r'\bun(\S{3,})'), r'(non|un|in|im)\g<1>', 2, 11),  # un
            (re.compile(r'\bnon(\S{3,})'), r'(non|un|in|im)\g<1>', 3, 11)]  # non


def get_medial_morphs():
    """
    list of tuple(regex, replacement string, count) for modifying 
        the middle of a term
    count represents the number of letters which are becoming
        optional (not _new_ optional ones)
    """
    return [(re.compile(r'\b(\S+)-(\S+)\b'), r'\g<1>(-| )?\g<2>', 1)]  # hyphen


def get_suffix_morphs():
    """
    list of tuple(regex, replacement string, count) for modifying 
        the end of a term
    count represents the number of letters which are becoming
        optional (not _new_ optional ones)
    """
    return [(re.compile(r'(\S{3,})ous\b'), r'\g<1>(ous)?', 3),  # -ous
            (re.compile(r'(\S{3,})al\b'), r'\g<1>(al|um)?', 2),  # -al
            (re.compile(r'(\S{1,})y\b'), r'\g<1>(y|ie)?', 1),  # -y pl.
            (re.compile(r'(\S{2,})oma\b'), r'\g<1>oma(ta|l|al)?', 0),  # -oma
            (re.compile(r'(\S{3,})ic\b'), r'\g<1>(ic)?', 2),  # -ic
            (re.compile(r'(\S{3,})oid\b'), r'\g<1>(oid|al)?', 3),  # -oid
            (re.compile(r'(\S{3,})ed\b'), r'\g<1>ed?', 1),  # -ed
            (re.compile(r'(\S{3,})ar\b'), r'\g<1>(ar)?', 2),  # -ar
            (re.compile(r'(\S{3,})is\b'), r'\g<1>(is|ic|es)?', 2),  # -is
            (re.compile(r'(\S{3,})ing\b'), r'\g<1>(ing|ed|e)?', 3),  # -ing
            (re.compile(r'(\S{3,})(s|t)ive\b'), r'\g<1>\g<2>(ive|ion)?', 3)]  # -sive


def get_final_morphs():
    """
    DEPRECATED -- if-elif-else expression more effective
        since difficult to sub at the end (after non-letters)
        
    Accounts for plurals.
    list of tuple(regex, replacement string, count) for computing
        the plural of a term
    count represents the number of letters which are becoming
        optional (not _new_ optional ones)
    """
    return [(re.compile(r'\b(\S+)s(\W*)\b'), r'\g<1>\g<2>s(es)?', 0),  # pl='es'
            (re.compile(r'\b(\S+)o(\W*)\b'), r'\g<1>\g<2>o(e?s)?', 0),  # pl='es'
            (re.compile(r'\b(\S+)(\W*)\b'), r'\g<1>\g<2>s?', 0)]  # pl='s'


def convert_to_regex(strings_ids_rxvar, convert_all=True,
                     affixmorphs=None, suffixmorphs=None,
                     medialmorphs=None, finalmorphs=None):
    """
    Takes a list of string-id tuples which are the words which
        need to be converted to regular expressions.
        The id part should be a unique identifier.
    Duplicate regexes will not be included. Duplicates
        are defined by whether or not a previous regex
        matches a succeeding one.
        
    Parameters:
        string_ids_settings: list of (string, id, wdOrder ) in which 
            string is to be converted to regex;
            order is important as the strings are
            converted to regexes in list order, and
            any following strings which are matched
            by the previous regexes will not be included
        degree of regex variation-
                    -1: no variation at all (no prefixes/suffixes)
                    0: no variation; words must be exact
                    1: minimal variation
                    2: moderate variation
                    3: flexible
        convert_all=True: if false, all strings will be
            converted to regexes regardless of a previous
            regex matching it
            :param strings_ids_rxvar:
            :param convert_all:
            :param affixmorphs:
            :param suffixmorphs:
            :param medialmorphs:
            :param finalmorphs:
    """
    if not affixmorphs:
        affixmorphs = get_affix_morphs()
    if not medialmorphs:
        medialmorphs = get_medial_morphs()
    if not suffixmorphs:
        suffixmorphs = get_suffix_morphs()

    regexes = []  # [(regex, id), ...]
    updated_ids = {}  # new_id -> old_id
    for string, id_, regex_variation in strings_ids_rxvar:
        string = string.replace('(', '').replace(')', '')
        string = string.replace('*', r'\w*')
        if regex_variation == -1:
            regexes.append((re.compile(r'\b{}\b'.format(string), re.I), id_))
            continue

        if not convert_all:
            # see if string is matched by previously-
            #    created RegEx
            found = False
            for rx, rx_id in regexes:
                if rx.match(string):
                    # print string,'duplicate of',rx,rx_id
                    found = True
                    if rx_id not in updated_ids:
                        updated_ids[rx_id] = {rx_id}
                    updated_ids[rx_id].add(id_)
                    break
            if found:
                continue  # Duplicate RegEx was found

        length = len(string)
        start_idx = 0  # ensure first letter is unable to change
        orig_string = string

        # 1) convert string to regex
        # a) convert any prefix
        for morph, repl, L, rLen in affixmorphs:
            if morph.match(string):
                string = morph.sub(repl, string)
                length -= L
                start_idx += rLen  # advance start index to letter after prefix
                break
        # b) convert any medial elements
        for morph, repl, L in medialmorphs:
            if morph.match(string):
                string = morph.sub(repl, string)
                length -= L
                break
        # c) convert any suffix
        for morph, repl, L in suffixmorphs:
            if morph.match(string):
                string = morph.sub(repl, string)
                length -= L
                break
        # d) add final elements
        if len(orig_string) > 2:
            if orig_string[-1] == 's':
                string += r'(es)?'
            elif orig_string[-1] == 'o':
                string += r'(e?s)?'
            else:
                string += r's?'

        # 2) add Error permissions
        length -= 1  # removing the first term
        # FLEXIBLE
        if regex_variation == 3:
            if length in [4, 5, 6]:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){i<2,d<2}')
            elif length in [7, 8, 9]:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){1i+1d<4}')
            elif length >= 10:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){e<4}')
        # MODERATE
        elif regex_variation == 2:
            if length in [4, 5, 6]:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){s<2}')
            elif length in [7, 8, 9]:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){1i+1d<3}')
            elif length >= 10:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){e<3}')
        # MINIMAL
        elif regex_variation == 1:
            if length in [4, 5, 6]:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){s<2}')
            elif length in [7, 8, 9]:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){1i+1d<2}')
            elif length >= 10:
                string = (string[:start_idx + 1] + '(' + string[start_idx + 1:] + '){e<2}')
        # NONE
        elif regex_variation == 0:
            pass

        # 3) add word boundary
        string = r'\b' + string + r'\b'

        # 4) compile string to regex
        rx = re.compile(string, re.V1 | re.I)
        # try:
        #     assert rx.match(orig_string)
        # except AssertionError:
        #     raise ValueError(r'Regex {} failed to match original string "{}".'.format(rx, orig_string))
        regexes.append((rx, id_))

    return regexes, updated_ids



"""

Edit:
2013-12-24    added line-ending splitting
2013-12-05    added begin/end offsets for conceptminer2
2013-11-26    added option for conceptminer2
"""

import argparse
import logging
import logging.config
from typing import List

from pytakes import Document, Dictionary, Output
from pytakes import MinerCollection, ConceptMiner, StatusMiner
from pytakes.parser import parse_processor
from pytakes import SentenceBoundary


def prepare(documents: List[Document], dictionaries: List[Dictionary],
            outputs: List[Output], negation_dicts: List[Dictionary]):
    """
    :param documents:
    :param dictionaries:
    :param outputs:
    :param negation_dicts:
    """
    mc = MinerCollection(ssplit=SentenceBoundary().ssplit)
    mc.add(ConceptMiner(dictionaries))
    mc.add(StatusMiner(negation_dicts))

    for out in outputs:
        out.create_output()

    logging.info('Retrieving notes.')
    for document in documents:
        for num, doc in enumerate(document.read_next()):
            if num % 100 == 0:
                logging.info(f'Completed: {num:>5}.')

            for sent_no, (sentence, cleaned_text) in enumerate(mc.parse(doc)):
                for feat, new in sentence:
                    for out in outputs:
                        out.writerow(doc.get_metalist(), feat, text=doc.get_text())
    logging.info('Completed: 100%')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-j', '--json-config',
                        help='Json file containing configuration information.')
    args = parser.parse_args()
    parse_processor(args.json_config)


if __name__ == '__main__':
    main()



from ..nlp.collections import MinerCollection


class TextItem(object):
    """ Carries metainformation and text for a document
    """

    def __init__(self, text, **meta):
        self.meta = meta
        if isinstance(text, str):
            self._orig_text = [text]
        else:
            self._orig_text = [t for t in text if t]
        try:
            self.text_ = self.fix_text(text[0])
        except IndexError as e:
            self.text_ = ''
        for txt in text[1:]:
            self.add_text(txt)  # split added 20131224

    def get_orig_text(self):
        return self._orig_text[0]

    def add_text(self, text):
        self.text_ += '\n' + self.fix_text(text)

    def get_text(self):
        return self.text_

    def get_metalist(self):
        return self.meta

    def fix_text(self, text):
        text = ' '.join(text.split('\n'))
        text.replace('don?t', "don't")  # otherwise the '?' will start a new sentence
        return text


def process_textitem(ti: TextItem, mc: MinerCollection):
    for res, sent in mc.parse(ti):
        yield res, sent



from datetime import datetime

from pytakes import CsvDictionary, CsvOutput, JsonlOutput
from pytakes import ConceptMiner, SentenceBoundary, MinerCollection, StatusMiner
from pytakes.corpus import get_next_from_corpus
from pytakes.dict.textitem import process_textitem


def output_context_manager(outfile, **kwargs):
    if outfile.endswith('jsonl'):
        return JsonlOutput(outfile, **kwargs)
    elif outfile.endswith('csv'):
        return CsvOutput(outfile, **kwargs)
    else:
        raise ValueError(f'Unrecognized file type: {outfile}')


def load_keywords(*paths):
    return ConceptMiner([CsvDictionary(**file) for file in paths])


def load_negation(version=None, path=None, skip=False, variation=0):
    if skip:
        return None
    elif path:  # requirement: must precede version as sometimes 'path' is always specified
        return StatusMiner(path=path, rx_var=variation)
    elif version:
        return StatusMiner(tablename=f'status{version}', rx_var=variation)
    else:
        raise ValueError('No negation cues provided!')


def run(input_dir, output_dir, *keyword_files, outfile=None, negex_version=1,
        negex_path=None, skip_negex=False, hostname=None):
    """

    :param hostname: this is mostly for testing so that the output won't be machine-dependent
    :param skip_negex: don't run negex
    :param input_dir:
    :param output_dir:
    :param keyword_files:
    :param outfile:
    :param negex_version:
    :param negex_path: if included, will ignore table (will only use file)
    :return:
    """
    mc = MinerCollection(ssplit=SentenceBoundary().ssplit)
    mc.add(load_keywords(*({'path': path} for path in keyword_files)))
    mc.add(load_negation(negex_version, negex_path, skip=skip_negex))
    if not outfile:
        outfile = 'extracted_concepts_{}.jsonl'.format(datetime.now().strftime('%Y%m%d_%H%M%S'))
    with output_context_manager(outfile, path=output_dir, metalabels=['file'], hostname=hostname) as out:
        for ti in get_next_from_corpus(input_dir):
            for results, sent in process_textitem(ti, mc):
                for result in results:
                    out.writerow(result, meta=list(ti.meta.values()), text=sent)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('-i', '--input-dir', dest='input_dir', required=True)
    parser.add_argument('-o', '--output-dir', dest='output_dir', required=True)
    parser.add_argument('-k', '--keyword-files', nargs='+', dest='keyword_files', required=True)
    parser.add_argument('--outfile', dest='outfile', default=None)
    parser.add_argument('--negex-version', dest='negex_version', default=1, type=int)
    parser.add_argument('--skip-negex', dest='skip_negex', action='store_true', default=False)
    parser.add_argument('--negex-path', dest='negex_path', default=None,
                        help='Specify csv file to use for negation rather than default.')
    args = parser.parse_args()
    run(args.input_dir, args.output_dir, *args.keyword_files,
        outfile=args.outfile, negex_version=args.negex_version,
        skip_negex=args.skip_negex, negex_path=args.negex_path)



import logging
import platform

import pyodbc
from jinja2 import Template

from . import templates
from .base import Dictionary, Output, Document
from ..dict.textitem import TextItem


class SqlDictionary(Dictionary):

    def __init__(self, dbi=None, schema=None, valence=None,
                 regex_variation=None, word_order=None, **kwargs):
        super().__init__(**kwargs)
        self.dbi = dbi
        self.valence = valence
        self.regex_variation = regex_variation
        self.word_order = word_order
        if schema:
            self.fullname = '{}.{}'.format(schema, self.name)

    def read(self):
        return self._get_terms()

    def _get_terms(self):
        """
        Retrieve terms from table.
        Function checks to see if optional columns are present,
        otherwise uses cTAKES defaults.
        """
        logging.info('Getting Terms.')
        # if [dbo] or [MASTER\...] prefaced to tablename
        columns = (x.lower() for x in self.dbi.get_table_columns(self.name))

        try:
            return self.dbi.execute_fetchall(Template(templates.PROC_GET_TERMS).render({
                'columns': columns,
                'valence': self.valence,
                'regex_variation': self.regex_variation,
                'word_order': self.word_order,
                'term_table': self.fullname
            }))
        except pyodbc.ProgrammingError as pe:
            logging.exception(pe)
            logging.error('Ensure that the term table has the variables "id", "text", and "cui".')
            raise pe


class SqlDocument(Document):

    def get_ids(self):
        """
        Retrieve documents from table (for batch mode)
        """
        self.dbi.execute(Template(templates.PROC_GET_DOC_IDS).render({
            'table_id': self.meta[0],
            'doc_table': self.fullname,
            'order_by': self.order_by
        }))
        return (x[0] for x in self.dbi)  # remove lists

    def __len__(self):
        pass

    def __init__(self, dbi=None, schema=None, where_clause=None,
                 order_by=None, batch_size=None, meta=None,
                 text=None, batch_mode=None, **config):
        super().__init__(**config)
        self.text = text
        self.meta = meta
        self.dbi = dbi
        if schema:
            self.fullname = '{}.{}'.format(schema, self.name)
        self.order_by = order_by
        self.batch_size = batch_size
        self.where_clause = '{} > {}'.format(self.meta[0], batch_mode)

    def read_next(self):
        """
        Retrieve documents from table
        """
        sql = Template(templates.PROC_GET_DOCS).render({
            'where_clause': self.where_clause,
            'order_by': self.order_by,
            'batch_size': self.batch_size,
            'meta_labels': self.meta,
            'text_labels': self.text,
            'doc_table': self.fullname
        })
        self.dbi.execute(sql)
        for row in self.dbi:
            yield TextItem(row[:-len(self.text)], row[-len(self.text):])


class SqlOutput(Output):

    def __init__(self, dbi=None, schema=None, labels=None,
                 types=None, hostname=None, batch_number=None, force=False, **config):
        super().__init__(**config)
        self.dbi = dbi
        self.labels = labels
        self.types = types
        self.hostname = hostname or platform.node()
        self.batch_number = batch_number
        self.force = force
        self.fullname = self.name
        if schema:
            self.fullname = '{}.{}'.format(schema, self.fullname)
        if batch_number:
            self.fullname = '{}_{}'.format(self.fullname, batch_number)

    def close(self):
        self.dbi.close()

    def delete_table_rows(self):
        """Drop all rows from destination table.

        :return: None
        """
        sql = "TRUNCATE TABLE {}".format(self.fullname)
        logging.debug(sql)
        self.dbi.execute_commit(sql)

    def create_output(self):
        try:
            self.dbi.execute_commit(Template(templates.PROC_CREATE_TABLE).render({
                'destination_table': self.fullname,
                'labels_types': zip(self.labels, self.types)
            }))
            logging.info('Table created: %s.' % self.fullname)
        except pyodbc.ProgrammingError as pe:
            logging.warning('Table already exists.')
            logging.error(pe)
            if self.force:
                logging.warning('Force deleting rows from table {}.'.format(self.fullname))
                self.delete_table_rows()
            else:
                logging.error('Add option "force" to delete rows from table.')
                raise pe
        except Exception as e:
            logging.exception(e)
            logging.error('Failed to create table.')
            raise e

    def writerow(self, meta, feat, text=None):
        self.dbi.execute_commit(
            Template(templates.PROC_INSERT_INTO2_QUERY).render(
                labels=self.labels, metas=meta,
                destination_table=self.fullname, feature=feat, text=text,
                captured=text[feat.begin():feat.end()].strip(),
                context=text[self._get_index(len(text), feat.begin() - 75):self._get_index(len(text), feat.end() + 75)],
                hostname=self.hostname, batch_number=self.batch_number
            )
        )
        self.dbi.execute_commit(
            Template(templates.PROC_INSERT_INTO3_QUERY).render(
                labels=self.labels, metas=meta,
                destination_table=self.fullname, feature=feat,
                hostname=self.hostname, batch_number=self.batch_number
            )
        )



"""Basic information extraction tool."""
__version__ = '1.0.5'

from .iolib import *
from .nlp import *
from .dict import TextItem

__all__ = [
    'CsvDictionary', 'CsvOutput', 'CsvDocument',
    'SqlDictionary', 'SqlOutput', 'SqlDocument',
    'SasDictionary', 'SasDocument',
    'JsonlOutput',
    'Dictionary', 'Output', 'Document',
    'TextItem',
    'Concept', 'Term', 'Word', 'Negation',
    'StatusMiner', 'MinerCollection', 'ConceptMiner',
    'SentenceBoundary'
]



import abc


class Miner(object):

    def __init__(self):
        pass

    @abc.abstractmethod
    def clean(self, terms):
        return terms

    @abc.abstractmethod
    def mine(self, text, offset):
        return text.split()

    @abc.abstractmethod
    def postprocess(self, terms):
        return terms

    @abc.abstractmethod
    def extract(self, terms):
        return list()



"""
Author: David Cronkite, GHRI
Date: 4apr13
Description:
    Utility functions for repeated NLP use.
"""

import regex as re
from datetime import datetime


def ssplit(texts):
    """ 
    Basic sentence splitting module which will also replace
    stray '~~' and '~' assumed as input for some pathology
    data.
    """
    ssplit_p = re.compile(r'(\S.+?[.!?\n])(?=\s+|$)')
    sections = []
    for text in texts:
        if not text: continue
        text = text.strip() + '\n'
        if text == '' or text == '\n': continue
        line = text.replace('\n', '\n ').replace('~~', '\n ')
        line = line.replace('~', ' ').replace('     ', ' ')
        line = line.replace('  ', ' ').replace('  ', ' ')
        result = re.findall(ssplit_p, line)
        if result:
            sections += [res.strip() for res in result]
        else:
            sections += [line.strip()]
    return sections


def psplit(texts):
    """
    Phrase splitting.
    """
    if isinstance(texts, str):
        texts = [texts]
    ssplit_p = re.compile(r'(\S.+?[.!?\n;:])(?=\s+|$)')
    sections = []
    for text in texts:
        if not text: continue
        text = text.strip() + '\n'
        if text == '' or text == '\n': continue
        line = text.replace('\n', '\n ').replace('~~', '\n ')
        line = line.replace('~', ' ').replace('     ', ' ')
        line = line.replace('  ', ' ').replace('  ', ' ')
        result = re.findall(ssplit_p, line)
        if result:
            sections += [res.strip() for res in result]
        else:
            sections += [line.strip()]
    return sections


def replace_punctuation(text):
    return re.sub(r"\p{P}+", "", text)


def is_number(s):
    try:
        float(s)
    except ValueError:
        return False
    return True


def convert_sql_datetime_to_python(datestring):
    return datetime.strptime(datestring, '%Y-%m-%d %H:%M:%S.%f')



"""
ghri.wildcat.conceptminer2.py
    created: 2013-04-18

Purpose:
    Mine concepts from the text. Essentially do what cTAKES does,
    but do it better AND simpler.

Author:
    Cronkite, David (GHRI)

Edits:
2013-11-05    added possible tags
2013-11-26    replaced by conceptminer2 for additional assertion annotation
-- conceptminer2 --
2013-12-12    added

"""
import copy
import numbers
from typing import List

from ..iolib.base import Dictionary
from . import convert
from .miner import Miner
from .terms import Term, Concept


class ConceptMiner(Miner):

    def __init__(self, dictionaries: List[Dictionary]):
        super().__init__()
        self.entries = [entry for d in dictionaries for entry in d.read()]
        self.cid_to_cat = {}  # ConceptID -> category
        self.cid_to_tids = {}  # ConceptID to TermIDs
        self.cid_search_param = {}
        self.wordID = 0
        self.tid_to_tid = {}  # list of conversions of
        # new TermIDs to old TermIDs
        # this should be a one-to-many
        # relationship where one New
        # TermID is equivalent to several
        # older TermIDs
        # no reason to have it the other
        # way around
        self.cid_to_val = {}
        self.cid_word_order = {}  # word order constraints
        self.wordlist = []

        self._unpack_concepts()

    def _unpack_concepts(self):
        """
        Organizes input from database
        Parameters:
            id_term_cat_val_rxVar_wdOrder - list of (id, term, category, valence,
                                       regex_variation, word_order)
                * id - number
                * term - ctakes dictionary "text" field
                * category - like s21 for sumres21 (DCIS)
                * valence - 0 if term contains a negation/uncertainty term
                * word_order- 0: free word order
                            1: enforce first word constraint
                            2: require precise word order
                * regex_variation-
                            0: no variation; words must be exact
                            1: minimal variation
                            2: moderate variation
                            3: flexible
        """
        for cid, term, cat, val, rxVar, wdOrder, max_intervening, max_search in self.entries:
            # update references of ConceptID (think "CUI")
            self.cid_to_cat[cid] = cat
            self.cid_to_val[cid] = val
            self.cid_word_order[cid] = wdOrder
            self.cid_search_param[cid] = (max_intervening, max_search)
            if wdOrder == 0:  # free word order
                self.cid_to_tids[cid] = set()
            elif wdOrder > 0:  # restricted word order
                self.cid_to_tids[cid] = list()

            for word in term.split():
                # give each word a unique id, and treated uniquely
                self.wordlist.append((word, self.wordID, rxVar))
                if wdOrder == 0:
                    self.cid_to_tids[cid].add(self.wordID)
                if wdOrder > 0:
                    self.cid_to_tids[cid].append(self.wordID)
                self.wordID += 1

        self.wordlist, upd_ids = convert.convert_to_regex(self.wordlist)

    def add_conversion(self, newtid_to_oldtids):
        """
        Adds new one-to-many relations between
        term_ids. Each term-id may only appear
        either on the RHS or the LHS of the dict
        (a.k.a., either keys or values, but not both)

        Parameters:
            newTid_to_oldTids -
                dictionary {newTid : set( [oldTid,oldTid,etc.])}
                where set(newTids) & set(oldTids) == set()
        """

        if not self.tid_to_tid:  # no extant conversions
            self.tid_to_tid = copy.deepcopy(newtid_to_oldtids)
        else:
            for newTid in newtid_to_oldtids:
                dest_tids = set()
                for oldTid in newtid_to_oldtids[newTid]:
                    dest_tids.add(newTid)
                    if oldTid in self.tid_to_tid:
                        dest_tids |= self.tid_to_tid[oldTid]
                        del self.tid_to_tid[oldTid]
                if newTid in self.tid_to_tid:
                    self.tid_to_tid[newTid] |= dest_tids
                else:
                    self.tid_to_tid[newTid] = dest_tids

    def get_original_term_id(self, term_ids):
        """
        Use the one-to-many mapping to get all possible term_ids for a particular term id (or list)

        @param term_ids: integer, or list of integers
        @return: all relevant term ids
        """
        if isinstance(term_ids, numbers.Real):
            term_ids = [term_ids]

        result = set(term_ids)
        for term_id in term_ids:
            if term_id in self.tid_to_tid:
                result |= self.tid_to_tid[term_id]
        return result

    def _check_valence(self, cid, judgment):
        """
        Checks the value of the term's valence.
        If valence==0, then the term must be negated in order to be positive.
            -e.g., 'hyperplasia without atypia' since 'without' will make
                    the entire phrase negative
        If valence==1, then the term is treated normally
            -e.g., 'hyperplasia with atypia'

        Return:
            True - should be negated
            False - should not be negated
        """
        if self.cid_to_val[cid]:  # is 1
            return judgment
        return 4 - judgment  # get inverse of certainty

    def _get_remaining(self, all_term_ids, curr_term_ids, word_order, first_word=False):
        """
        1. Check if there is an overlap between those terms desired by the current
            concept (all_term_ids) and the currently found term (curr_term_ids)
            if not, return None
        2. Get the remaining terms for the current concept, and return them

        @param all_term_ids:
        @param curr_term_ids:
        @param word_order: 0: free word order
                    1: enforce first word constraint
                    2: require precise word order
        @param first_word: True: current term is first word of potential concept
                     False: current term is in middle/end of potential concept
        @return: remaining terms for the current concept (empty list of all found), or None not found
        """
        if word_order == 0 or (word_order == 1 and not first_word):
            shared_set = (all_term_ids & curr_term_ids)
            if shared_set:
                remain_set = (curr_term_ids - shared_set)
                return remain_set
            else:
                return None
        elif word_order == 1 and first_word:
            if curr_term_ids[0] in all_term_ids:
                remain_set = set(curr_term_ids[1:])
                return remain_set
            else:
                return None
        elif word_order == 2:
            if curr_term_ids[0] in all_term_ids:
                remain_list = curr_term_ids[1:]
                return remain_list
            else:
                return None

    def extract(self, terms):
        """
        Aggregate terms (in the word list) into concepts according.

        @param terms: list of word-derived objects including negation, words, and terms
            only Terms will be considered in determining concepts; must be sorted
        @return: all found concepts

        """
        concepts = []
        for i in range(len(terms)):
            cword = terms[i]
            if isinstance(cword, Term):
                c_all_tids = set(self.get_original_term_id(cword.id()))
                for cid in self.cid_to_tids:  # look through concepts
                    max_intervening, max_search = self.cid_search_param[cid]
                    remain_set = self._get_remaining(c_all_tids, self.cid_to_tids[cid], self.cid_word_order[cid],
                                                     first_word=True)

                    if remain_set is None:
                        continue  # return type of None was not a match

                    # check if concept was completed
                    if remain_set:  # concept has additional terms (> 1 word)
                        concept = self._aggregate(terms[i:],
                                                  remain_set,
                                                  cword.get_certainty(),
                                                  cword.is_hypothetical(),
                                                  cword.is_historical(),
                                                  cword.is_not_patient(),
                                                  cword.begin(),
                                                  cid,
                                                  max_search,
                                                  max_intervening,
                                                  self.cid_word_order[cid])  # word order (added 2013-11-19)
                    else:  # one-term concept (remain_set is empty list/set)
                        concept = Concept(cword.word(),
                                          cword.begin(),
                                          cword.end(),
                                          cid,
                                          self.cid_to_cat[cid],
                                          self._check_valence(cid, cword.get_certainty()),
                                          cword.is_hypothetical(),
                                          cword.is_historical(),
                                          cword.is_not_patient(),
                                          offset=cword.offset_)
                    if concept:  # function might return "False"
                        concepts.append(concept)
            else:  # current word is not a Term
                continue
        return concepts

    def _aggregate(self, words, remain_set, certainty, hypothetical, historical,
                   not_patient, start_idx, cid, words_to_look_at, max_intervening_words, word_order):
        """
        Look through subsequent words for additional terms to determine if the concept is contained in the text.

        @param words: list of words and Terms (which compose concepts)
        @param remain_set: remaining words to complete concept
        @param certainty: negex status of first found term in concept
        @param hypothetical: negex status of first found term in concept
        @param historical: negex status of first found term in concept
        @param not_patient: negex status of first found term in concept
        @param start_idx: starting index of first word in concept
        @param cid: concept id of concept
        @param words_to_look_at: maximum number of words to look at
        @param max_intervening_words: maximum allowed number of intervening words between words in concept
        @param word_order: word ordering requirements for current concept
        @return: found Concept or False
        """
        words_to_look_at_incr = words_to_look_at
        orig_words = words
        words = words[1:]
        # see if matching terms are available in the next couple terms
        for j in range(len(words)):
            if j < words_to_look_at and max_intervening_words >= 0:
                nword = words[j]
                if isinstance(nword, Term):
                    # check if current term is present in current concept
                    n_all_tids = set(self.get_original_term_id(nword.id()))
                    temp_remain_set = self._get_remaining(n_all_tids, remain_set, word_order, first_word=False)

                    if temp_remain_set is None:  # term not in concept
                        max_intervening_words -= 1
                    else:  # term in concept
                        words_to_look_at += words_to_look_at_incr
                        # update negex status with more egregious form
                        certainty = min(certainty, nword.get_certainty())
                        hypothetical = max(hypothetical, nword.is_hypothetical())
                        historical = max(historical, nword.is_historical())
                        not_patient = max(not_patient, nword.is_not_patient())
                        if temp_remain_set:  # more terms to find
                            remain_set = temp_remain_set
                        else:  # empty list or set (cannot be None)
                            return Concept(' '.join([w.word() for w in orig_words[:j + 2]]),
                                           start_idx,
                                           words[j].end(),
                                           cid,
                                           self.cid_to_cat[cid],
                                           self._check_valence(cid, certainty),
                                           hypothetical,
                                           historical,
                                           not_patient,
                                           offset=nword.offset_)

            else:
                break
        return False

    def postprocess(self, terms):
        return terms  # no postprocessing

    def mine(self, text, offset):
        """
        Paramters:
            terms - list of (term,id) where unique id for each term
            :param text:
            :param offset:
        """
        results = []
        for term, id_ in self.wordlist:
            for m in term.finditer(text):
                results.append(Term(term, m.start(), m.end(), 'term', id_, offset=offset))
        return sorted(results)

    def clean(self, terms):
        """
        remove terms that are subsumed by other terms
        :param terms:
        """
        terms.sort(reverse=False)
        if len(terms) < 2:
            return terms
        i = 1
        curr = 0
        while True:
            if terms[curr] == terms[i]:
                if len(terms[curr]) > len(terms[i]):
                    del terms[i]
                elif len(terms[curr]) < len(terms[i]):
                    del terms[curr]
                    curr = i - 1
                elif terms[curr].begin() == terms[i].begin() and terms[curr].end() == terms[i].end():
                    terms[curr].add_term(terms[i])
                    del terms[i]
                else:  # keep both representations
                    curr = i
                    i += 1
            else:
                curr = i
                i += 1

            if i >= len(terms):
                return terms



import os

base_path = r'BASE_PATH'

config = {
    'corpus': {  # how to get the text data
        'directories': [  # specify path to .txt files
            os.path.join(base_path, 'data', 'files')
        ],
    },
    'keywords': [  # path to keyword files, usually stored as CSV
        {
            'path': os.path.join(base_path, 'data', 'concepts.csv')
        }
    ],
    'negation': {
        'path': os.path.join(base_path, 'data', 'negation.csv'),
    },
    'output': {
        'path': os.path.join(base_path, 'data', 'testout'),
        'outfile': 'run.negex.jsonl',  # name of output file (or given default name)
        'hostname': 'Eumaeus'
    },
}

print(config)



from .utils import ssplit, psplit
from .conceptminer import ConceptMiner
from .terms import Concept, Term, Word, Negation
from .statusminer import StatusMiner
from .collections import MinerCollection
from .sentence_boundary import SentenceBoundary



import collections


def remaining_time(complete, total, start_time, end_time, format_='%Y-%m-%d %H:%M:%S,%f'):
    """
    Utility function for determining average length of time for each unit.

    Parameters:
        complete- int, how many of total have been completed
        total - int, total number of instances to be completed
        start_time - time as string when process started
        end_time - time as string when process completed
        format - format in which times are encoded
    """
    from datetime import datetime, timedelta
    percent = float(total) / float(complete)
    start = datetime.strptime(start_time, format_)
    end = datetime.strptime(end_time, format_)
    diff = end - start
    total_time = timedelta(seconds=int(percent * diff.total_seconds()))
    return str(total_time - diff)


def recursively_output_dictionary(d, out, padding='\t'):
    """
    Recursively list out an embedded dictionary. Write
    the result to out with padding for each level of
    embeddedness.

    Parameters:
        d - embedded dictionary
        out - output file ready for writing, or sys.stdout, etc.
        padding - amount of new indentation for each level of embedding
        :param d:
        :param out:
        :param padding:
    """
    out.write(recursive_dict_to_string(d, padding))


def recursive_dict_to_string(d, padding='\t'):
    """
    Recursively list out an embedded dictionary. Return as string.

    Parameters:
        d - embedded dictionary
        padding - amount of new indentation for each level of embedding

    Return:
        string representation of embedded dictionary
        :param d:
        :param padding:
    """
    return '\n'.join(_recursive_dict_to_string(d, padding))


def _recursive_dict_to_string(d, padding, curr_pad=''):
    """
    Helper method for recursive_dict_to_string containing recursive aspect.

    Parameters:
        d - embedded dictionary
        padding - amount of new indentation for each level of embedding
        curr_pad - the amount of padding for the current depth

    Return:
        list representation of embedded dictionary
        each item represents a newline.
    """
    if isinstance(d, dict):
        results = []
        for key in sorted(d):
            if isinstance(d[key], dict):
                results += [curr_pad + str(key)]
                results += _recursive_dict_to_string(d[key], padding, curr_pad + padding)
            else:
                results += [curr_pad + str(key) + padding + str(d[key])]

        return results
    else:
        return [str(d)]


def get_valid_args(func, dct):
    argnames = func.__code__.co_varnames[:func.__code__.co_argcount]
    return dict((key, val) for key, val in list(dct.items()) if key in argnames)


def flatten(it):
    for x in it:
        if (isinstance(x, collections.Iterable) and
                not isinstance(x, str)):
            yield from flatten(x)
        else:
            yield x



NEGATION_TYPE = 'negn'
IMPROBABLE_TYPE = 'impr'
POSSIBLE_TYPE = 'poss'
PROBABLE_TYPE = 'prob'
HYPOTHETICAL_TYPE = 'hypo'
FUTURE_TYPE = 'futp'
HISTORICAL_TYPE = 'hist'
OTHER_SUBJECT_TYPE = 'othr'

CERTAINTY = 'certainty'
HYPOTHETICAL = 'hypothetical'
HISTORICAL = 'historical'
OTHER_SUBJECT = 'otherSubject'



config = {
    'corpus': {  # how to get the text data
        'directories': [  # specify path to .txt files
            r'PATH'
        ],
        'connections': [  # specify other connection types
            {
                'name': 'TABLENAME',
                'name_col': 'TEXT ID COLUMN',
                'text_col': 'TEXT COLUMN',
                # specify either driver/server/database OR connection_string
                # connection string examples here: https://docs.sqlalchemy.org/en/13/core/engines.html
                'connection_string': 'SQLALCHEMY-LIKE CONNECTION_STRING',
                # db args: driver/server/database
                'driver': 'DRIVER',  # available listed in pytakes/iolib/sqlai.py, or use connection string
                'server': 'SERVER',
                'database': 'DATABASE',
            }
        ]
    },
    'keywords': [  # path to keyword files, usually stored as CSV
        {
            'path': r'PATH',
            'regex_variation': 0  # set to -1 if you don't want any expansion
        }
    ],
    'negation': {  # select either version or path (not both)
        'path': r'PATH TO NEGATION CSV FILE',
        'version': 1,  # int (version number), built-in/default
        'skip': False,  # bool: if skip is True: don't do negation
    },
    'output': {
        'path': r'PATH TO OUTPUT DIRECTORY',
        'outfile': 'NAME.out.jsonl',  # name of output file (or given default name)
        'hostname': ''
    },
}

print(config)



import copy
import os
from itertools import product

import regex as re

CUI_IDX, CAT_IDX, CONFIG_IDX = range(3)


class Config(list):
    def __init__(self, *args):
        if args:
            args = args[0]
        newiter = []
        for el in args:
            newiter.append([el])
        list.__init__(self, newiter)

    def add(self, lst):
        for idx, el in enumerate(lst):
            if self.__len__() > idx:
                self[idx] = self[idx] + el
            else:
                self.append(el)

    def get(self, idx, default, func=min):
        if self.__len__() > idx:
            try:
                res = func(x for x in self.__getitem__(idx) if x is not None)
            except ValueError:  # min() on empty list
                return default
            return res
        return default


def generate_combinations(rules, cats, generator):
    rows = []
    for cui, rule, configs in rules:
        if not cui:
            cui = generator.generate_cui()
        if not configs:
            configs = Config()
        cui = cui[:8]  # limit cui length to 8 characters
        lst = []
        start = 0
        for m in re.finditer(r'\[(?P<CATEGORY>\w+?)\]', rule):
            if m.start() > start:
                lst.append([rule[start: m.start()]])
            lst.append(cats[m.group('CATEGORY')])
            start = m.end()
        # final fencepost
        if start < len(rule) - 1:
            lst.append([rule[start:]])
        rows.append((cui, list(product(*lst)), configs))
    return rows


def build_rows(rows, generator):
    results = []
    for cui, c_rows, configs in rows:
        for row in c_rows:
            row_cui, row_configs, text = get_meta_for_row(cui, configs, row)
            # cui, fword, term, textlength, code, regexVariation, wordOrder
            insert = [None] * 8
            insert[0] = row_cui
            insert[2] = ' '.join((''.join(text).strip()).split())  # term
            insert[1] = insert[2].split(' ')[0]  # firstword
            # noinspection PyTypeChecker
            insert[3] = len(insert[2])  # textlength
            insert[4] = generator.generate_code()  # unique code
            # min is the most conservative parameter (in the following...)
            insert[5] = row_configs.get(0, default=1, func=min)  # regexVariation
            # max is the most conservative parameter in the following...
            insert[6] = row_configs.get(1, default=1, func=max)  # wordOrder
            # valence doesn't do anything yet...it needs to be 1
            insert[7] = row_configs.get(2, default=1, func=max)  # valence
            results.append(insert)
    return results


def get_meta_for_row(cui, configs, row):
    """
    get the min of each config
    NB: cui is ignored, and returns only the given CUI
        - I can't figure out how to make sense of
        word-level CUIs...which trumps the other?
    """
    text = []
    row_configs = copy.copy(configs)

    for el in row:
        if isinstance(el, (list, tuple)):
            row_cui, content, conf = el
            row_configs.add(conf)
            text.append(content)
            if not cui or cui.startswith('G'):
                cui = row_cui
        else:
            text.append(el)

    return cui, row_configs, text


#  RULES
def read_rules(rulefile, generator):
    rules = []
    with open(rulefile) as f:
        for line in f.readlines():
            line = line.strip()

            # remove comments
            if not line or line[0] == '#':
                continue
            elif '#' in line:
                line = line[:line.index('#')].strip()
            cui, rule, configs = parse_rule(line)
            generator.used_cui(cui)
            rules.append((cui, rule, configs))
    return rules


def parse_rule(line):
    # get optional cui
    if '==' in line:
        cui, rule = line.split('==')
        cui.strip()
    else:
        rule = line
        cui = None

    # get optional configs ## added v1.1
    if '%%' in rule:
        rule, configs = rule.split('%%')
        configs = Config([int(x) if x.strip() else None for x in configs.split(',')])
    else:
        configs = Config()
    return cui, rule.strip(), configs


#  CATEGORIES
def read_files(catfiles, categories):
    for file in catfiles:
        read_file(file, categories)
    return categories


def read_file(catfile, categories):
    with open(catfile) as f:
        cat = os.path.splitext(os.path.split(catfile)[-1])[0]
        for line in f.readlines():
            line = line.strip()
            
            # remove comments
            if not line or line[0] == '#':
                continue
            elif '#' in line:
                line = line[:line.index('#')].strip()

            m = re.match(r'\[(\w+?)]', line)
            if m:  # found category name
                cat = m.group(1)
                if cat not in categories:
                    categories[cat] = []
                continue

            else:  # found names belonging to category
                names = read_cat(line)
                if cat not in categories:
                    categories[cat] = []
                categories[cat] += names


def read_cat(cats):
    """
    cats - list of lines in category file,
        name(||name2||...)%%config1,config2

    NB: cui== is ignored!!
    """
    names = []
    for cat in cats if isinstance(cats, list) else [cats]:
        cat = cat.strip()
        cui, cat, configs = parse_rule(cat)
        # divide line into separate synonyms
        names += [(cui, c.strip(), configs) for c in cat.split('||')]
    return names



import os
import csv
import platform

from .base import Dictionary, Output, Document
from ..dict.textitem import TextItem
from ..nlp.terms import Concept


class CsvDictionary(Dictionary):

    def __init__(self, name=None, path=None, valence=1, regex_variation=0, word_order=1,
                 max_intervening=1, max_search=2, **kwargs):
        super().__init__(name=name, **kwargs)
        self.valence = valence
        self.regex_variation = regex_variation
        self.word_order = word_order
        self.max_intervening = max_intervening
        self.max_search = max_search
        if name and path:
            self.fp = os.path.join(path, self.name)
        elif name:
            self.fp = self.name
        else:
            self.fp = path
            self.name = os.path.basename(path)

    def read(self):
        return self._get_terms()

    def _get_terms(self):
        """
        Retrieve terms from table.
        Function checks to see if optional columns are present,
        otherwise uses cTAKES defaults.
        """
        columns = {}
        res = []
        with open(self.fp) as fh:
            for i, line in enumerate(csv.reader(fh)):
                if i == 0:
                    columns = {x.lower(): i for i, x in enumerate(line)}
                else:
                    res.append((
                        line[columns['id']],
                        line[columns['text']],
                        line[columns['cui']],
                        self.int_or_default(line[columns.get('valence', self.valence)], self.valence),
                        self.int_or_default(line[columns.get('regexvariation', self.regex_variation)],
                                            self.regex_variation),
                        self.int_or_default(line[columns.get('wordorder', self.word_order)], self.word_order),
                        self.int_or_default(line[columns.get('maxintervening', self.max_intervening)],
                                            self.max_intervening),
                        self.int_or_default(line[columns.get('maxwords', self.max_search)], self.max_search),
                    ))
        return res


class CsvDocument(Document):

    def get_ids(self):
        pass

    def __len__(self):
        pass

    def __init__(self, path=None, order_by=None, batch_size=None, meta=None,
                 text=None, **config):
        super().__init__(**config)
        self.text = text
        self.meta = meta
        self.order_by = order_by
        self.batch_size = batch_size
        self.fp = os.path.join(path, self.name)

    def read_next(self):
        """
        Retrieve documents from table
        """
        columns = {}
        with open(self.fp) as fh:
            for i, line in enumerate(csv.reader(fh)):
                if i == 0:
                    columns = {x: i for i, x in enumerate(line)}
                else:
                    yield TextItem(
                        meta_list=[line[columns[m]] for m in self.meta],
                        text=[line[columns[t]] for t in self.text]
                    )


class CsvOutput(Output):

    def __init__(self, name=None, path=None, metalabels=None,
                 types=None, hostname=None, batch_number=None, **config):
        super().__init__(name=name, **config)
        self.labels = (metalabels or []) + self.all_labels
        self.types = types
        self.hostname = hostname or platform.node()
        self.batch_number = batch_number
        self.fp = os.path.join(path, self.name)
        self.fh = open(self.fp, 'w', newline='')
        self.writer = csv.writer(self.fh)
        self.writer.writerow(self.labels)  # header

    def writerow(self, feat: Concept, meta=None, text=None):
        if not meta:
            meta = []
        if text:
            length = len(text)
            self.writer.writerow(meta +
                                 [feat.id(),
                                  feat.cat(),
                                  self._get_text(text, feat.begin(), feat.end()),
                                  self._get_text(text, self._get_index(length, feat.begin() - self.context_width),
                                                 self._get_index(length, feat.end() + self.context_width)),
                                  feat.get_certainty(),
                                  feat.is_hypothetical(),
                                  feat.is_historical(),
                                  feat.is_not_patient(),
                                  feat.get_absolute_begin(),
                                  feat.get_absolute_end(),
                                  self.hostname,
                                  self.batch_number
                                  ]
                                 )
        else:  # no text
            self.writer.writerow(meta +
                                 [text[feat.begin():feat.end()].strip(),
                                  None,
                                  None,
                                  None,
                                  None,
                                  None,
                                  self.hostname,
                                  self.batch_number
                                  ]
                                 )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fh.close()



import json

import jsonschema


JSON_SCHEMA = {
    'type': 'object',
    'properties': {
        'corpus': {
            'type': 'object',
            'properties': {
                'directories': {
                    'type': 'array',
                    'items': {'type': 'string'}
                },
                'connections': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'name': {'type': 'string'},
                            'driver': {'type': 'string'},
                            'server': {'type': 'string'},
                            'database': {'type': 'string'},
                            'name_col': {'type': 'string'},
                            'text_col': {'type': 'string'}
                        }
                    }
                },
            }
        },
        'keywords': {  # paths to keyword files
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'path': {'type': 'string'},
                    'regex_variation': {'type': 'integer'},  # -1 to 3
                    'word_order': {'type': 'integer'},
                    'max_search': {'type': 'integer'},
                    'max_intervening': {'type': 'integer'},
                }
            }
        },
        'negation': {
            'type': 'object',
            'properties': {
                'version': {'type': 'integer'},  # built-in version
                'path': {'type': 'string'},
                'skip': {'type': 'boolean'},
                'variation': {'type': 'integer'},
            }
        },
        'output': {
            'type': 'object',
            'properties': {
                'outfile': {'type': 'string'},
                'path': {'type': 'string'},
                'hostname': {'type': 'string'},
            }
        },
        'logger': {
            'type': 'object',
            'properties': {
                'verbose': {'type': 'boolean'}
            }
        }
    }
}


def myexec(code):
    import warnings
    warnings.warn('Executing python external file: only do this if you trust it')
    import sys
    from io import StringIO
    temp_stdout = sys.stdout
    sys.stdout = StringIO()
    try:
        # try if this is a expression
        ret = eval(code)
        result = sys.stdout.getvalue()
        if ret:
            result = result + ret
    except:
        try:
            exec(code)
        except:
            # you can use <traceback> module here
            import traceback
            buf = StringIO()
            traceback.print_exc(file=buf)
            error = buf.getvalue()
            raise ValueError(error)
        else:
            result = sys.stdout.getvalue()
    sys.stdout = temp_stdout
    return result


def load_yaml(fh):
    try:
        import yaml  # pyyaml
    except ModuleNotFoundError:
        try:
            from ruamel import yaml
        except ModuleNotFoundError:
            raise ModuleNotFoundError('Missing module: `yaml`. Install pyyaml or ruamel.yaml from pip.')
    return yaml.load(fh)


def get_config(path):
    with open(path) as fh:
        if path.endswith('json'):
            return json.load(fh)
        elif path.endswith('yaml'):
            return load_yaml(fh)
        elif path.endswith('py'):
            return eval(myexec(fh.read()))
        else:
            raise ValueError('Unrecognized configuration file type: {}'.format(path.split('.')[-1]))


def validate_config(path):
    conf = get_config(path)
    jsonschema.validate(conf, JSON_SCHEMA)
    return conf



"""

"""
import json

from pytakes.iolib.config import get_data_item
from pytakes import MinerCollection
from pytakes.util.db_reader import DbInterface

DEFAULT_NEGATION_VARIATION = -1


def _parse(dicts, res, conn):
    ds = []
    for name in dicts:
        resource = res[dicts[name]['resource']]
        ds.append(
            get_data_item(
                resource,
                name,
                conn=conn[resource['connection']]
            )
        )
    return ds


def parse_miner(dicts, res, conn):
    miners = MinerCollection()
    for name in dicts:
        c = dicts[name]
        if c['type'] == 'keyword':
            for neg in c.get('negation', list()):
                negvar = neg.get('variation', DEFAULT_NEGATION_VARIATION)
                cxcn = res[c['resource']]
        elif c['type'] == 'ngram':
            pass
        else:
            raise ValueError('Unrecognized miner type: {}'.format(dicts[name]['type']))
    return miners


def parse_processor(json_file):
    config = json.load(json_file)
    res = config['resources']
    conns = config['connections']
    conn = {c: DbInterface(driver=c['driver'],
                           server=c['server'],
                           database=c['database']) for c in conns}
    dicts = _parse(config['dictionary'], res, conn)
    docs = _parse(config['document'], res, conn)
    out = _parse(config['output'], res, conn)
    miners = parse_miner(config['miner'], res, conn)
    return {
        'dictionary': dicts,
        'document': docs,
        'output': out
    }


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('--json-config',
                        help='')
    args = parser.parse_args()
    parse_processor(args.json_config)



"""
downloaded from http://code.google.com/p/negex/downloads/detail?name=negex.python.zip
on 9 July 2012

General NegEx Python Implementation v.1.2 (Peter Kang) & De-Identified Annotations (Chapman)

Edits:
    2012-07-09  - added isNegated and isAffirmed boolean methods
                - added return statement to __str__ method
    2012-12-10  - added sortRules2
    2013-04-17  - added sortRulesFromTuple (for database integration)
                - added myNegTagger class (customization of negTagger)
    2013-11-25  - added myStatusTagger class (expands myNegTagger for use with historical, etc.)
    2013-12-09  - added getNegex and getContext functions; I'm going to need these each
                time I use negex, so might as well include them
"""
import csv
import sqlite3

import pkg_resources
import regex as re

from .miner import Miner
from .terms import *


# noinspection PyShadowingNames
class StatusMiner(Miner):
    """
    Customizations of Peter Kang & Wendy Chapman's statusminer.py algorithm.

    Output Terms with negation/possibility flags inherent in the term
    rather than words.
    Terms are sorted by relative position in the sentence.

    Changes from myNegTagger:
        1. Addition of history of tags.
        2. Each tag has choice of direction


    Overview of use:
        1. initialize with rules (grab from db with sortRulesFromTuples)
        2. call findNegation on the text, and hold onto the return
            Negation(Term) objects
        3. convert all other words in the text to some sort of Term
            objects
        4. call negateSentence(s) to determine which Terms are
            negated (and, optionally, which terms are possible)
    """

    '''
    Defining types of tags
    '''
    STOP_TAGS = ['conj']  # end of all scopes
    NEGATION_TAGS = ['affm', 'prob', 'poss', 'impr', 'negn', 'pseu']
    HYPOTHETICAL_TAGS = ['hypo', 'indi']
    TEMPORAL_TAGS = ['futp', 'hist']
    SUBJECT_TAGS = ['subj', 'othr']
    NEGEX_LEVEL = {
        0: [(0, '')
            ],
        1: [(12, '{1i+1d<3})'),
            (8, '{1i+1d<2})'),
            (0, '')
            ],
        2: [(14, '{1i+1d<4}'),
            (10, '{1i+1d<3}'),
            (6, '{1i+1d<2}'),
            (0, '')
            ],
        3: [(12, '{1i+1d<4}'),
            (8, '{1i+1d<3}'),
            (4, '{1i+1d<2}'),
            (0, '')
            ]
    }

    def __init__(self, rules=None, rx_var=0, maxscope=100, tablename='status1',
                 path=None):
        """
        Parameters:
            rules - list of negation trigger terms from the sortRules function
                if None, use built-in status
            rxVar - allowable regular expression variation
                    0: no variation; words must be exact => default
                    1: minimal variation
                    2: moderate variation
                    3: flexible
            maxscope - maximum distance allowed for negation/etc.
                DEFAULt is 100 (i.e., unlimited)
        """
        super().__init__()
        self._rules = []
        self._maxscope = maxscope
        if rules is None:
            if path:
                rules = self.load_negex_from_csv(path)
            elif tablename:
                rules = self.load_negex_from_db(tablename)
            else:  # tablename was explicitly negated
                rules = []
        self.add_rules(rules, rx_var)

    @staticmethod
    def load_negex_from_db(tablename):
        conn = sqlite3.connect(pkg_resources.resource_filename('pytakes.nlp', 'data/status.db'))
        cur = conn.cursor()
        cur.execute(f'select negex, type, direction from {tablename}')
        for negex, _type, direction in cur:
            yield negex, _type.strip('[]'), int(direction)

    @staticmethod
    def load_negex_from_csv(path):
        with open(path, newline='') as fh:
            reader = csv.DictReader(fh)
            for row in reader:
                yield row['negex'], row['type'].strip('[]'), int(row['direction'])

    def add_rules(self, rules, rx_var):

        def get_negation_string(values, length):
            for x, y in values:
                if x > length:
                    return y
            return ''

        # add rules, but permit errors based on length
        for negex, _type, direction in self.sort_rules_for_status(rules):
            negex_pat = r'\b({})\b{}'.format(r'\W+'.join(negex.split()),
                                             get_negation_string(self.NEGEX_LEVEL[rx_var], len(negex)))
            self._rules.append((negex, re.compile(negex_pat, re.I), _type.strip('[]'), direction))

    @staticmethod
    def sort_rules_for_status(rulelist, exclusions=None):
        """
        Return sorted list of rules: (negex, type, direction, pattern)

        Input: list of tuples (negex, type, direction)

        Sorts list of rules descending based on length of rule,
        and converts the pattern into a regular expression.

        For use with myStatusTagger
        :param rulelist:
        :param exclusions:
        """
        sortedlist = []
        for negex, type_, direction in sorted(rulelist, key=lambda x: len(x[0]), reverse=True):
            if exclusions and negex in exclusions:
                continue
            sortedlist.append((negex, type_.strip('[]'), direction))
        return sortedlist

    def extract(self, terms):
        return []

    def mine(self, text, offset):
        """
        Find negations in a piece of text. If called sentence by
        sentence, set 'offset' to the len() of all previous
        sentences so that the Negation indices are correct.

        Parameters:
            text - piece of text to find negation
            offset - (default 0) len(all previous sentences)
                    if putting in only one sentence at a time
                    (not recommended as sentence boundaries
                    are not important for this portion of the
                    algorithm)
                    :return:
                    :param text:
                    :param offset:
        """
        negations = []
        # rules already sorted by length of negation expression
        for negex, pattern, _type, direction in self._rules:
            for m in pattern.finditer(text):
                n = Negation(negex, m.start(), m.end(), _type=_type, direction=direction, offset=offset)
                # longer sequences will trump smaller ones
                if n not in negations:
                    negations.append(n)
        return negations

    # noinspection PyPep8Naming
    def postprocess(self, terms):
        """
        :param terms:

        """
        # indices
        TYPE = 0
        STATUS = 1
        OVERLAP = 2
        COUNTER = 3

        def found_term(ind, _type):
            """ updates ind[icator] when a term is found
            :param ind:
            :param _type:
            """
            ind[TYPE] = _type
            ind[STATUS] = 1
            ind[OVERLAP] = 0
            ind[COUNTER] = 0

        def overlap_term(ind):
            """ updates ind[icator] when overlap found
            :param ind:
            """
            ind[OVERLAP] = 1

        def analyze_direction(sentence, directions):
            """
            directions - list of directions to look for
            1- backward
            2- forward
            3- both (bidirectional)
            :type sentence: list
            :param sentence:
            :param directions:
            """
            negtype = ['', 0, 0, 0]
            hypotype = ['', 0, 0, 0]
            temptype = ['', 0, 0, 0]
            subjtype = ['', 0, 0, 0]
            types = [negtype, hypotype, temptype, subjtype]
            for idx, term in enumerate(sentence):

                if term.direction() in directions:
                    if term.type() in self.NEGATION_TAGS:
                        found_term(negtype, term.type())
                    elif term.type() in self.HYPOTHETICAL_TAGS:
                        found_term(hypotype, term.type())
                    elif term.type() in self.TEMPORAL_TAGS:
                        found_term(temptype, term.type())
                    elif term.type() in self.SUBJECT_TAGS:
                        found_term(subjtype, term.type())

                elif term.type() in self.STOP_TAGS:
                    for _type in types:
                        overlap_term(_type)

                elif term.type() in self.NEGATION_TAGS:
                    overlap_term(negtype)

                elif term.type() in self.TEMPORAL_TAGS:
                    overlap_term(temptype)

                elif term.type() in self.HYPOTHETICAL_TAGS:
                    overlap_term(hypotype)

                elif term.type() in self.SUBJECT_TAGS:
                    overlap_term(subjtype)

                else:
                    for _type in types:
                        _type[COUNTER] += 1

                # check if term should be updated with any types
                for _type in types:
                    if _type[STATUS] == 1 and _type[OVERLAP] == 0 and _type[COUNTER] <= self._maxscope:
                        # ignoring affirmation tag (affm)
                        if _type[TYPE] == 'negn':
                            term.negate()
                        elif _type[TYPE] == 'impr':
                            term.improbable()
                        elif _type[TYPE] == 'poss':
                            term.possible()
                        elif _type[TYPE] == 'prob':
                            term.probable()
                        if _type[TYPE] == 'hypo':
                            term.hypothetical()
                        if _type[TYPE] == 'futp':
                            term.hypothetical()
                        if _type[TYPE] == 'hist':
                            term.historical()
                        if _type[TYPE] == 'othr':
                            term.other_subject()

            return sentence
            # end inner function

        # check all FORWARD (direction=2 or 3)
        sentence = analyze_direction(terms, [2, 3])

        # reverse and check all BACKWARD/BIDIRECTIONAL (1 or 3)      
        sentence.reverse()  # reverse sentence
        sentence = analyze_direction(terms, [1, 3])

        # put sentence back in correct order
        sentence.reverse()  # sentence correctly ordered

        return sentence

    def clean(self, text):
        return text



from datetime import datetime
from typing import List

from pytakes import MinerCollection, SentenceBoundary
from pytakes.corpus import get_next_from_corpus
from pytakes.dict.textitem import process_textitem
from pytakes.run.schema import validate_config
from pytakes.run.simple_run import load_keywords, load_negation, output_context_manager


def run(corpus=None, output=None, keywords: List = None, negation=None, logger=None):
    """

    :return:
    """
    mc = MinerCollection(ssplit=SentenceBoundary().ssplit)
    mc.add(load_keywords(*keywords))
    mc.add(load_negation(**negation))
    default_output = {
        'outfile': 'extracted_concepts_{}.jsonl'.format(datetime.now().strftime('%Y%m%d_%H%M%S')),
        'metalabels': ['file'],
    }
    with output_context_manager(**{**default_output, **output}) as out:
        for i, ti in enumerate(get_next_from_corpus(**corpus)):
            for results, sent in process_textitem(ti, mc):
                for result in results:
                    out.writerow(result, meta=list(ti.meta.values()), text=sent)


def run_config(config_file):
    run(**validate_config(config_file))



"""
automate_run.py

Author: Cronkite, David

PURPOSE
-----------
Automate creation of a directory to run pyctakes in batches.

TODO
-------


CHANGELOG
---------------
2013-12-12        created

"""

import argparse
import logging
import logging.config
import math
import os

from jinja2 import Template
from pytakes.util import mylogger
from pytakes.util.utils import get_valid_args

from pytakes import templates
from pytakes.util.db_reader import DbInterface


def get_integer(s):
    num = None
    while True:
        num = input(s)
        try:
            num = int(num)
            break
        except ValueError as e:
            print("Not a valid number.")
    return num


def is_this_okay():
    response = input('Is this okay? ')
    return 'y' in response.lower()


def get_batch_size(count):
    while True:
        batchsize = get_integer('Size of batches: ')
        batches = int(math.ceil(float(count) / batchsize))
        print("This will result in %d batches." % batches)
        if is_this_okay():
            return batchsize, batches


def get_number_of_files(batches):
    while True:
        filecount = get_integer('Number of batch files: ')
        filebatchcount = int(math.ceil(float(batches) / filecount))
        print('This will result in %d batches per file.' % filebatchcount)
        if is_this_okay():
            return filecount, filebatchcount


def resolve_formatting(label, value):
    if value:
        if isinstance(value, bool):
            return '--{}'.format(label)
        elif isinstance(value, list):
            return '--{}\n{}'.format(label, '\n'.join(value))
        else:
            return '--{}={}'.format(label, value)
    else:
        return ''


def create_batch_file(output_dir, batch_label, document_table, destination_table,
                      batch_size, batch_start, batch_end, driver, server, database,
                      meta_labels, primary_key, text_labels,
                      options, python, pytakes_path, tracking_method, send_email):
    with open(os.path.join(output_dir, 'pytakes-batch{}.bat'.format(batch_label)), 'w') as out:
        if pytakes_path:
            out.write(Template(templates.RUN_BATCH_FILE).render(
                batch_number=batch_label, python=python, pytakes_path=pytakes_path, send_email=send_email))
        else:
            out.write(Template(templates.RUN_COMMAND_BATCH_FILE).render(
                batch_number=batch_label, send_email=send_email
            ))

    if primary_key in meta_labels:
        meta_labels.remove(primary_key)

    with open(os.path.join(output_dir, 'pytakes-batch{}.conf'.format(batch_label)), 'w') as out:
        out.write(
            Template(templates.RUN_CONF_FILE).render(
                driver=driver, server=server, database=database, document_table=document_table,
                destination_table=destination_table,
                meta_labels=meta_labels, primary_key=primary_key, text_labels=text_labels,
                options=options,
                batch_size=batch_size, batch_start=batch_start, batch_end=batch_end,
                tracking_method=tracking_method
            ))


def create_email_file(output_dir, filecount, destination_table,
                      recipients, sender, mail_server_address, python, pytakes_path):
    with open(os.path.join(output_dir, 'email.conf'), 'w') as out:
        out.write(
            Template(templates.EMAIL_CONF_FILE).render(
                recipients=recipients, filecount=filecount, destination_table=destination_table,
                sender=sender, mail_server_address=mail_server_address
            ))

    with open(os.path.join(output_dir, 'bad_email.conf'), 'w') as out:
        out.write(
            Template(templates.BAD_EMAIL_CONF_FILE).render(
                recipients=recipients, filecount=filecount, destination_table=destination_table,
                sender=sender, mail_server_address=mail_server_address
            ))


def create_post_process_batch(pp_dir, destination_table, negation_table, negation_variation, driver,
                              server, database, batch_count, python, pytakes_path,
                              tracking_method, send_email):
    with open(os.path.join(pp_dir, 'postprocess.bat'), 'w') as out:
        if pytakes_path:
            out.write(Template(templates.PP_BATCH_FILE).render(
                pytakes_path=pytakes_path, python=python
            ))
        else:
            out.write(templates.PP_COMMAND_BATCH_FILE)
    with open(os.path.join(pp_dir, 'postprocess.conf'), 'w') as out:
        out.write(Template(templates.PP_CONF_FILE).render(
            driver=driver, server=server, database=database,
            destination_table=destination_table, negation_table=negation_table,
            negation_variation=negation_variation, batch_count=batch_count,
            python=python, pytakes_path=pytakes_path, tracking_method=tracking_method))


def automate_run(dbi, cm_options, concept_miner, document_table,
                 output_dir, destination_table,
                 driver, server, database,
                 meta_labels, primary_key, text_labels,
                 recipients, sender, mail_server_address, negation_table, negation_variation,
                 python, pytakes_path,
                 tracking_method):
    count = dbi.fetch_rowcount(document_table)
    logging.info('Found %d documents in %s.' % (count, document_table))
    batchsize, batchcount = get_batch_size(count)
    filecount, batchesperfile = get_number_of_files(batchcount)

    logging.info('Number of batches: %d' % batchcount)
    logging.info('Batch size: %d' % batchsize)
    logging.info('Number of files: %d' % filecount)
    logging.info('Batches per file: %d' % batchesperfile)

    options = [resolve_formatting(x, y) for x, y in cm_options if y]
    if not primary_key:
        primary_key = meta_labels[0]
    elif primary_key in meta_labels:
        meta_labels.remove(primary_key)

    send_email = recipients and sender and mail_server_address
    os.makedirs(output_dir, exist_ok=True)
    if send_email:
        create_email_file(output_dir, filecount, destination_table,
                          recipients, sender, mail_server_address, python, pytakes_path)
    else:
        logging.warning(
            'Skipping email: Failed to specify at least one of "recipients", "sender", and "mail server address".')

    postprocess_dir = os.path.join(output_dir, 'post')

    batch_start = 1
    for batch_label in range(1, filecount + 1):
        batch_end = batch_start + batchesperfile
        create_batch_file(output_dir, batch_label, document_table, destination_table,
                          batchsize, batch_start, batch_end, driver, server, database,
                          meta_labels, primary_key, text_labels,
                          options, python, pytakes_path, tracking_method, send_email)
        batch_start = batch_end

    if concept_miner == 2:
        os.makedirs(postprocess_dir, exist_ok=True)
        create_post_process_batch(postprocess_dir, destination_table, negation_table, negation_variation,
                                  driver, server, database, filecount + 1, python, pytakes_path,
                                  tracking_method, send_email)
    logging.info('Completed.')


def main():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('--driver', default='SQL Server', help='Driver to connect with.')
    parser.add_argument('-s', '--server', required=True, help='Database server to use.')
    parser.add_argument('-d', '--database', required=True, help='Database to use.')
    parser.add_argument('--document-table', help='Table of input documents.')
    parser.add_argument('--output-dir', help='Destination directory.')
    parser.add_argument('--destination-table', help='Output table, to be created.')

    parser.add_argument('--concept-miner', default=2, type=int, required=False,
                        help='Version of concept miner to use.')
    # concept miner 2
    parser.add_argument('--max-intervening-terms', default=1, type=int,
                        help='Max number of terms that can occur between searched for terms.')
    parser.add_argument('--max-length-of-search', required=False, default=3, type=int,
                        help='Max number of words in which to look for the next term.')
    parser.add_argument('--valence', required=False, default=None, type=int)
    parser.add_argument('--regex-variation', required=False, default=-1, type=int)
    parser.add_argument('--word-order', required=False, default=None, type=int)
    parser.add_argument('--dictionary-table', help='Term table/dictionary table as input.')
    parser.add_argument('--negation-table', help='Negation/status table with (negex,status,direction).')
    parser.add_argument('--negation-variation', help='Negation variation [0-3].')

    # concept miner 3
    parser.add_argument('--stopwords', required=False, default=None, nargs='+',
                        help='Stopwords to include.')
    parser.add_argument('--number-normalization', required=False, action='store_true', default=False,
                        help='Whether or not to normalize all numbers to a default value.')
    parser.add_argument('--stopword-tables', required=False, default=None, nargs='+',
                        help='Tables containing relevant stopwords. NYI')

    parser.add_argument('--meta-labels', required=True, nargs='+',
                        help='Identifying labels to include in output.')
    parser.add_argument('--primary-key', required=False, default=None,
                        help='Primary key for documents. This will be used in sorting batches. '
                             'If not included, the first meta label will be the primary key.')
    parser.add_argument('--text-labels', nargs='+', help='Name of text columns.')

    parser.add_argument('--tracking-method', choices=['name', 'column'], default='name',
                        help='Method to track progress of batches.')

    # library -- these are more for debugging
    parser.add_argument('--python', default='python', help='Specify Python version.')
    parser.add_argument('--pytakes-path', default='',
                        help='Path to pytakes directory, which should contain the "src" directory.'
                             ' Default assumes that you have installed this file in scripts.')

    parser.add_argument('-v', '--verbosity', type=int, default=2, help='Verbosity of log output.')

    # email
    parser.add_argument('--recipients', required=False, default=[], nargs='+',
                        help='In format of "name,email@address"')
    parser.add_argument('--sender', required=False, default=None,
                        help='In format of "name,email@address"')
    parser.add_argument('--mail-server-address', required=False,
                        help='Mail server address.')

    # autogenerate sample config file
    cparser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    cparser.add_argument('--create-sample', action='store_true', default=False,
                         help='Generate a sample configuration file to fill out.')

    args, _ = cparser.parse_known_args()
    if args.create_sample:
        with open(args.create_sample, 'w') as out:
            out.write(templates.SAMPLE_CONF_FILE)
        return

    args = parser.parse_args()

    loglevel = mylogger.resolve_verbosity(args.verbosity)
    logging.config.dictConfig(mylogger.setup(name='automate_run', loglevel=loglevel))

    dbi = DbInterface(driver=args.driver, server=args.server, database=args.database, loglevel=loglevel)

    if args.concept_miner == 2:
        cm_options = [
            ('concept-miner', 2),
            ('term-table', args.dictionary_table),
            ('max-intervening-terms', args.max_intervening_terms),
            ('max-length-of-search', args.max_length_of_search),
            ('valence', args.valence),
            ('regex-variation', args.regex_variation),
            ('word-order', args.word_order),
            ('destination-table', '{}_pre'.format(args.destination_table))
        ]
        if args.negation_table:
            cm_options += [('negation-table', args.negation_table),
                           ('negation-variation', args.negation_variation)]

    elif args.concept_miner == 3:
        cm_options = [
            ('concept-miner', 3),
            ('stopwords', args.stopwords),
            ('number-normalization', args.number_normalization),
            ('stopword-tables', args.stopword_tables),
            ('destination-table', args.destination_table)
        ]
    else:
        raise ValueError('Invalid argument for concept miner.')

    # clean recipients (see PYT-15)
    args.recipients = [x for x in args.recipients if x and '@' in x]

    try:
        automate_run(dbi, cm_options, **get_valid_args(automate_run, vars(args)))
    except Exception as e:
        logging.exception(e)
        logging.error('Process terminated with errors.')
    logging.info('Process completed.')


if __name__ == '__main__':
    main()



from pytakes import StatusMiner, Word, Negation, Term


def test_negation_bidirectional():
    miner = StatusMiner(rules=[])
    terms = [
        Term('I', 0, 1, 'C01', 1),
        Word('have', 2, 6),
        Negation('no', 7, 9, 'negn', 3),
        Term('beer', 10, 14, 'C02', 2),
    ]
    miner.postprocess(terms)
    assert terms[0].is_negated()
    assert terms[-1].is_negated()


def test_negation_forward():
    miner = StatusMiner(rules=[])
    terms = [
        Term('I', 0, 1, 'C01', 1),
        Word('have', 2, 6),
        Negation('no', 7, 9, 'negn', 2),
        Term('beer', 10, 14, 'C02', 2),
    ]
    miner.postprocess(terms)
    assert not terms[0].is_negated()
    assert terms[-1].is_negated()


def test_negation_backward():
    miner = StatusMiner(rules=[])
    terms = [
        Term('I', 0, 1, 'C01', 1),
        Word('have', 2, 6),
        Negation('no', 7, 9, 'negn', 1),
        Term('beer', 10, 14, 'C02', 2),
    ]
    miner.postprocess(terms)
    assert terms[0].is_negated()
    assert not terms[-1].is_negated()


def test_negation_with_punct():
    miner = StatusMiner(rules=[('c/o', 'negn', 3)])
    assert len(miner.mine('pt c/o', 0)) == 1
    assert len(miner.mine('pt c / o', 0)) == 0


def test_negation_with_punct2():
    miner = StatusMiner(rules=[('c / o', 'negn', 3)])
    assert len(miner.mine('pt c/o', 0)) == 0
    assert len(miner.mine('pt c / o', 0)) == 1


def test_negation_with_punct3():
    miner = StatusMiner(rules=[(r'c\s*/\s*o', 'negn', 3)])
    assert len(miner.mine('pt c/o', 0)) == 1
    assert len(miner.mine('pt c / o', 0)) == 1


def test_negation_end_scope_conj():
    miner = StatusMiner(rules=[])
    terms = [
        Term('I', 0, 1, 'C01', 1),
        Word('have', 2, 6),
        Negation('no', 7, 9, 'negn', 3),
        Term('beer', 10, 14, 'C02', 2),
        Negation('or', 15, 17, 'conj', 0),
        Term('wine', 18, 22, 'C02', 3),
    ]
    miner.postprocess(terms)
    assert terms[0].is_negated()
    assert terms[-3].is_negated()  # negates 'beer'
    assert not terms[-1].is_negated()  # 'or' blocks 'wine' from being negated



"""
Run script for executing pyTAKES. Relies on config file defined according to pytakes/schema.py
"""
from pytakes.run.runner import run_config

if __name__ == '__main__':
    import sys

    run_config(sys.argv[1])



"""
Create a negex-styled dictionary from a tab-delimited
negation list, and a negation equivalences file.

"""

import argparse
import copy
import logging
import logging.config
import pyodbc

import regex as re
from .util import mylogger

from pytakes.util.db_reader import DbInterface


def parse_equivalence_file(eqfile, equivs):
    with open(eqfile) as f:
        for line in f.readlines():
            equiv, defn = line.strip().split('=')
            if equiv in equivs:
                equivs[equiv] += defn.split('||')
            else:
                equivs[equiv] = defn.split('||')


def parse_equivalence_files(eqfiles):
    equivs = {}  # { equiv : list(synonym, synonym, etc.) }
    for eqFile in eqfiles:
        parse_equivalence_file(eqFile, equivs)
    return equivs


def write_termlist_to_file(termlist, outfile, columns):
    with open(outfile, 'w') as out:
        out.write('\t'.join(columns) + '\n')
        for result in termlist:
            if len(columns) == 3:
                terms, _type, direction = result
                out.write('\t'.join([terms, _type, direction]) + '\n')
            else:
                raise Exception('This should not happen.')
    return True


def write_termlist_to_db(termlist, dbi, table_name, columns):
    sqltypes = ['varchar(255)', 'varchar(10)', 'int']
    create_query = ','.join([col + ' ' + st for col, st in zip(columns, sqltypes)])
    try:
        dbi.execute_commit('''
            CREATE TABLE %s
            (
                %s
            );
        ''' % (table_name, create_query))
    except pyodbc.ProgrammingError:
        logging.warning('Appending to existing table "%s".' % table_name)

    insert_query = ','.join(columns)
    values_query = None
    try:
        for result in termlist:
            values_query = ','.join([x if c == 'int' else "'" + x + "'" for x, c in zip(result, columns) if c])
            dbi.execute_commit('''
                INSERT INTO %s (%s)
                VALUES (%s)
            ''' % (table_name, insert_query, values_query))
    except Exception as e:
        logging.exception(e)
        logging.error('Unable to insert values: "%s".' % values_query)

    return True


def create_negex(negex_files, eqfiles, output_file=None, dbi=None, table_name='test',
                 instances=('negex', 'type', 'direction')):
    length_of_instances = 0
    equivs = parse_equivalence_files(eqfiles)
    final_termlist = []
    pattern = re.compile('%([A-Z-_]+)%([A-Za-z-_]*)')
    for negexfile in negex_files:
        with open(negexfile) as f:
            for num, line in enumerate(f.readlines()):
                line = line.strip().split('#')[0]
                if not line or line[0] == '#':
                    continue
                lst = line.split('\t')
                length_of_instances = len(lst)
                if len(lst) < 3:
                    logging.warning('Insufficient instances in line %d of %s. Skipping.' % (num, negexfile))
                    logging.debug(lst)
                    continue
                elif len(lst) == 3:
                    term, _type, direction = lst
                else:
                    logging.warning('Only the first three negation configurations are being read (%s).' % negexfile)
                    term, _type, direction = lst[:3]

                results = [([], _type, direction)]  # start with a single empty list
                for tt in term.split():
                    m = pattern.match(tt)
                    if m:
                        new_results = []
                        for termlist, _type, direction in results:
                            synonyms = equivs[m.group(1)]
                            for eq in synonyms:
                                ctermlist = copy.copy(termlist)
                                ctermlist.append(eq + m.group(2))
                                new_results.append((ctermlist, _type, direction))
                        results = new_results
                    else:
                        for termlist, _type, direction in results:
                            termlist.append(tt)
                final_termlist += [(' '.join(x), y, z) for x, y, z in results]

    if output_file:
        write_termlist_to_file(final_termlist, output_file, instances[:length_of_instances])

    if dbi:
        write_termlist_to_db(final_termlist, dbi, table_name, instances[:length_of_instances])


def main():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('--negex-files', required=True, nargs='+')
    parser.add_argument('--equiv-files', required=True, nargs='+')
    parser.add_argument('--output-file', required=False, default=False,
                        help='If included, negexes will write to file.')
    parser.add_argument('--output-driver', required=False, default=False,
                        help='Driver for output table to be written.')
    parser.add_argument('--output-server', required=False, default=False,
                        help='Server for output table to be written.')
    parser.add_argument('--output-database', required=False, default=False,
                        help='Database for output table to be written.')
    parser.add_argument('--output-table', required=False, default=False,
                        help='If included negexes will write to table.')

    parser.add_argument('--verbosity', type=int, default=2, help='Verbosity of log output.')
    args = parser.parse_args()

    if args.verbosity <= 0:
        loglevel = logging.ERROR
    elif args.verbosity == 1:
        loglevel = logging.WARNING
    elif args.verbosity == 2:
        loglevel = logging.INFO
    else:
        loglevel = logging.DEBUG

    logging.config.dictConfig(mylogger.setup(loglevel=loglevel))

    if args.output_table:
        dbi = DbInterface(args.output_driver, args.output_server, args.output_database)
    else:
        dbi = None

    create_negex(args.negex_files, args.equiv_files, args.output_file, dbi, args.output_table)


if __name__ == '__main__':
    main()



"""

TODO: Make into sql_alchemy.
"""

import pyodbc
import time
import logging


def connection_continue(func):
    def wrapper(*args, **kwargs):
        while True:
            try:
                return func(*args, **kwargs)
            except pyodbc.Error as e:
                args[0].wait_for_connection_resume(60, e)
    return wrapper


class DbInterface(object):
    def __init__(self, driver='', server='', database='', loglevel=None):
        """
        open connection
        connection will be closed when it falls out of scope

        Parameters
        -----------
        database - specify database to connect to
        loglevel - deprecated; no longer has any effect
        """
        self._cs = 'DRIVER={%s};SERVER=%s;DATABASE=%s;' % (driver, server, database)
        self.conn = pyodbc.connect(self._cs)
        self.cur = self.conn.cursor()
        # default loglevel is 30 (i.e., logger.WARNING)
        self.logger = logging.getLogger(__name__)
        if loglevel:
            self.logger.setLevel(loglevel)

    def close(self):
        self.__del__()

    def __del__(self):
        """close connection when garbage collecting"""
        try:
            self.conn.close()
        except AttributeError:
            pass

    @connection_continue
    def __next__(self):
        return next(self.cur)

    @connection_continue
    def __iter__(self):
        return self

    @connection_continue
    def fetchall(self):
        return self.cur.fetchall()

    @connection_continue
    def fetchmany(self):
        return self.cur.fetchmany()

    @connection_continue
    def fetchone(self):
        return self.cur.fetchone()

    @connection_continue
    def execute(self, text, debug=False):
        if debug:
            print(text)
        logging.debug(text)
        self.cur.execute(text)

    @connection_continue
    def commit(self):
        self.conn.commit()

    @connection_continue
    def execute_commit(self, text, debug=False):
        self.execute(text, debug)
        self.commit()

    @connection_continue
    def execute_return(self, text, debug=False):
        return self.execute_fetchall(text, debug)

    @connection_continue
    def execute_fetchone(self, text, debug=False):
        self.execute(text, debug)
        return self.fetchone()

    @connection_continue
    def execute_fetchmany(self, text, debug=False):
        self.execute(text, debug)
        return self.fetchmany()

    @connection_continue
    def execute_fetchall(self, text, debug=False):
        self.execute(text, debug)
        return self.fetchall()

    @connection_continue
    def fetch_rowcount(self, table, debug=False):
        sql = '''SELECT COUNT(*) FROM {}'''.format(table)
        return self.execute_fetchone(sql, debug)[0]

    def get_table_columns(self, table_name):
        """returns columns of specified table """
        if self.is_sql_server_connection():
            cols = self.execute_fetchall('''
                SELECT column_name
                FROM INFORMATION_SCHEMA.COLUMNS
                WHERE table_name = '%s'
            ''' % table_name)
            return (x[0] for x in cols)
        else:
            raise ValueError('Unsupported database connection.')

    def is_sql_server_connection(self):
        return 'sql server' in self._cs.lower()

    # noinspection PyUnboundLocalVariable
    def wait_for_connection_resume(self, waittime, err):
        while True:
            if isinstance(err, pyodbc.ProgrammingError):
                if err.args[0] == '42000' and '(6005)' in err.args[1]:
                    logging.warning(err)
                else:
                    raise err
            elif isinstance(err, pyodbc.OperationalError):
                if err.args[0] == 'HYT00' and '(0)' in err.args[1]:
                    logging.warning(err)
                else:
                    raise err
            else:  # isinstance(err, pyodbc.Error):
                if err.args[0] == '08001' and '(17)' in err.args[1]:
                    logging.warning(err)
                elif err.args[0] == '01000' and 'General network error' in err.args[1]:
                    logging.warning(err)
                elif err.args[0] == '42000' and '(4060)' in err.args[1]:
                    logging.warning(err)
                else:
                    raise err
            logging.info('Waiting for connection resume.')
            time.sleep(waittime)
            try:
                self.conn = pyodbc.connect(self._cs)
                self.cur = self.conn.cursor()
                break
            except pyodbc.Error as err:
                continue


def parse_dbargs(args):
    import argparse

    db_parser = argparse.ArgumentParser(fromfile_prefix_chars='@')

    db_parser.add_argument('--password', metavar='pwd', required=True, help='SQL Server password.')
    db_parser.add_argument('--user', metavar='u', default=None, help='SQL Server username.')
    db_args, _ = db_parser.parse_known_args(args)
    return db_args



import os

import pytest

from pytakes import CsvDictionary
from pytakes import MinerCollection
from pytakes import ConceptMiner
from pytakes import StatusMiner


@pytest.fixture
def path():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def concept_file(path):
    return os.path.join(path, r'data\concepts.csv')


def test_statusminer_negated():
    s = StatusMiner()
    negations = s.mine('no evidence of anything', 0)
    assert len(negations) == 1


def test_statusminer_hypothetical():
    s = StatusMiner()
    negations = s.mine('if he wants to', 0)
    assert len(negations) == 1


def test_negated_sentence(concept_file):
    mc = MinerCollection()
    mc.add(ConceptMiner([CsvDictionary(concept_file)]))
    mc.add(StatusMiner())
    concepts, sentence = mc.parse_sentence('no bad weather', offset=0)
    for concept in concepts:
        assert concept.is_negated()


def test_improbable_sentence(concept_file):
    mc = MinerCollection()
    mc.add(ConceptMiner([CsvDictionary(concept_file)]))
    mc.add(StatusMiner())
    concepts, sentence = mc.parse_sentence('no evidence of bad weather', offset=0)
    for concept in concepts:
        assert concept.is_improbable()


def test_hypothetical_sentence(concept_file):
    mc = MinerCollection()
    mc.add(ConceptMiner([CsvDictionary(concept_file)]))
    mc.add(StatusMiner())
    concepts, sentence = mc.parse_sentence('if there is bad weather', offset=0)
    for concept in concepts:
        assert concept.is_hypothetical()



"""
 # <p>Title: Sentence Boundary program </p>
 # <p>Create Date: 16:00:49 03/21/11</p>
 # <p>Copyright: Copyright (c) Department of Biomedical Informatics </p>
 # <p>Company: Vanderbilt University </p>
 # @author Yonghui Wu
 # @version 1.2
 # <p>Description: This program is used to detect sentence boundary and format each sentence as a single line </p>
 # Input: 1: English_dictionary 2: Abbreviation_dictionary 3: input_dir 4: output_dir
 # Output: Each file in the file_list will be processed by this sentence boundary tool,
 the out put will be generated in output directory with name: "file_name.sent"

Modified so that it will:
    1. read input from database
    2. parse sentences as parameters (a string/list) and return the result
    3. this will require a class-like interface
    
Edited: 2may13
Editor: David Cronkite, GHRI
"""
import re
import sqlite3
import pkg_resources


class SentenceBoundary(object):
    def __init__(self, ignorecase=True, config=None, debug=False):
        """

        :param ignorecase: case insensitive (default=True)
        :param config: dictionary which may contain the following keys
        .. note::
            prep -- list of prepositions
            det -- list of determiners
            conj -- list of conjunctions
            non_stop_punct -- list of stop punctuation
            sentence_word -- list of sentence words
            knuthus -- list of knuthus words
            abbrevs -- list of abbreviation words

        :param debug: enable debug mode (default=False)
        :return:
        """
        self.prep = {'about', 'above', 'across', 'after', 'against', 'aka', 'along', 'and', 'anti', 'apart', 'around',
                     'as', 'astride', 'at', 'away', 'because', 'before', 'behind', 'below', 'beneath', 'beside',
                     'between', 'beyond', 'but', 'by', 'contra', 'down', 'due to', 'during', 'ex', 'except',
                     'excluding', 'following', 'for', 'from', 'given', 'in', 'including', 'inside', 'into', 'like',
                     'near', 'nearby', 'neath', 'of', 'off', 'on', 'onto', 'or', 'out', 'over', 'past', 'per', 'plus',
                     'since', 'so', 'than', 'though', 'through', 'til', 'to', 'toward', 'towards', 'under',
                     'underneath', 'versus', 'via', 'where', 'while', 'with', 'within', 'without', 'also'}
        self.det = {'a', 'an', 'the'}
        self.conj = {'and', 'or', 'but', 'if', 'nor', 'for', 'except', 'although', 'no'}
        self.non_stop_punct = {}
        self.sentence_word = {'we', 'us', 'patient', 'denies', 'reveals', 'no', 'none', 'he', 'she', 'his', 'her',
                              'they', 'them', 'is', 'was', 'who', 'when', 'where', 'which', 'are', 'be', 'have', 'had',
                              'has', 'this', 'will', 'that', 'the', 'to', 'in', 'with', 'for', 'an', 'and', 'but', 'or',
                              'as', 'at', 'of', 'have', 'it', 'that', 'by', 'from', 'on', 'include'}

        if config and 'knuthus' in config and 'abbrevs' in config:
            self.knuthus = {}
            self.abbrevs = {}
        else:
            conn = sqlite3.connect(pkg_resources.resource_filename('pytakes.nlp', 'data/ssplit.db'))
            cur = conn.cursor()
            cur.execute('SELECT abbr FROM abbreviations')
            self.abbrevs = {x[0] for x in cur.fetchall()}
            cur.execute('SELECT word FROM knuthus')
            self.knuthus = {x[0] for x in cur.fetchall()}

        for el in config or []:
            setattr(self, el, self.load_set_lower(config[el]))

        self.abbrevs -= self.knuthus  # remove duplicates
        self.ignorecase = ignorecase
        self.debug = debug

    def ssplit(self, text, debug=False):
        # debug outputs print statements
        debug_ = (debug or self.debug)

        # sentences = []
        text = self.prepare_text(text)
        i = -1
        while i + 1 < len(text):
            i += 1
            txt = text[i]
            words = txt.split()
            tmp = ''
            for j, word in enumerate(words):
                pos = self.has_dot(word)
                if pos >= 0:
                    dot_num = self.num_dot(word)
                    if dot_num == 1:
                        if self.is_stop_punct(word):  # single dot
                            tmp += word + '\n'
                        # '.' begining, do not change
                        elif pos == 0:
                            tmp += word + ' '
                        # end of word
                        elif pos == len(word) - 1:
                            if self.is_num_list(word):
                                if j == 0:  # 1. Percocet 5/325 one p.o.
                                    tmp += word + ' '
                                else:  # postoperative day 3.
                                    tmp += word[:-1] + ' .\n'
                            # afebrile .
                            elif (word[:-1].lower() not in self.abbrevs and
                                    word.lower() not in self.abbrevs):
                                tmp += (word[:-1] + ' .\n')
                            else:  # abbreviations
                                if j + 1 < len(words):
                                    nword = words[j + 1]
                                else:
                                    nword = ''
                                # elevation MI. The patient was
                                if (len(nword) > 0 and
                                        self.isupper(nword[0], True) and
                                    # 1.0 MM. C) LEFT BREAST, ...
                                    # added by djc 30jul13
                                        (nword.lower() in self.sentence_word or
                                            (2 <= len(nword) <= 3 and
                                             nword[0].isalpha() and
                                             nword[-1] == ')'
                                             )
                                         )):
                                    tmp += word + '\n'
                                else:
                                    tmp += word + ' '
                        else:
                            lword, rword = word.split('.')
                            if (self.isupper(rword[0], True) and
                                    (rword in self.sentence_word or
                                     rword.lower() in self.knuthus)):
                                tmp += lword + '.\n' + rword + ' '
                            else:
                                tmp += word + ' '

                    else:
                        # "CK-MB of 9.1. His ..."
                        if word[-1] == '.':
                            if self.is_digit(word[:-1]):
                                tmp += word[:-1] + ' .\n'
                            else:  # "without need for O.T. or P.T.  He is"
                                if j + 1 < len(words):
                                    nword = words[j + 1]
                                else:
                                    nword = ''

                                if (len(nword) > 0 and
                                        self.isupper(nword[0], True) and
                                        nword.lower() in self.sentence_word):
                                    tmp += word + '\n'
                                else:
                                    tmp += word + ' '
                        else:
                            tmp += word + ' '

                else:  # normal word
                    tmp += word + ' '

            # end for j, word
            tmp = tmp.strip()
            if i + 1 < len(text):
                words = text[i + 1].split()
                nword = words[0]
                words = tmp.split()
                eword = words[-1].lower()
                # handle 'Page : '
                if ((len(txt) > 6 and
                        txt[:6].lower() == 'page :') or
                        (len(text[i + 1]) > 6 and
                            text[i + 1].lower() == 'page :')):
                    text[i] = (tmp.strip() + '\n')
                elif tmp[-1] == '.':
                    if (len(tmp) >= 2 and
                            tmp[-2] != ' ' and
                            not (nword == '-' or
                                 self.is_num_list(nword) or
                                 self.isupper(nword[0]))):
                        text[i] = (tmp.strip() + ' ')
                    else:
                        text[i] = (tmp.strip() + '\n')
                elif text[i][-1] != '.' and txt[-1] != ':':
                    # ends in prep, conj, det, and non_stop_punct
                    if (eword in self.prep or eword in self.det or
                            eword in self.non_stop_punct or eword in self.conj):
                        text[i] = (tmp.strip() + ' ')
                    elif not (nword == '-' or
                              self.is_num_list(nword) or
                              self.isupper(nword[0])):
                        text[i] = (tmp.strip() + ' ')
                    else:
                        text[i] = (tmp.strip() + '\n')
                else:
                    text[i] = (tmp.strip() + '\n')

            else:
                text[i] = tmp
        # end for i, txt
        if text:
            text[-1] += '\n'
        for group in text:
            for sentence in group.split('\n'):
                if sentence.strip():
                    yield sentence

    def isupper(self, word, default=False):
        if self.ignorecase:
            return default
        return word.isupper()

    def load_set_lower(self, lst):
        """
        Load dictionary in to a set, convert into lowercase.
        :param lst:
        """
        sett = set()
        for line in lst:
            line = line[0].strip()
            if len(line) > 0:
                sett.add(line.lower())
        return sett

    def num_dot(self, word):
        """
        return the number of dot in word
        :param word:
        """
        i = 0
        for w in word:
            if w == '.' or word[i] == '!' or word[i] == '?' or word[i] == ';':
                i += 1
        return i

    def has_dot(self, word):
        """
        Return the position of '.' in a word. -1 denote there are no '.' appeared
        :param word:
        """
        i = 0
        lenn = len(word)
        while i < lenn:
            if word[i] == '.' or word[i] == '!' or word[i] == '?' or word[i] == ';':
                return i
            i += 1
        return - 1

    def remove_double_star(self, line):
        """
        Some auto - inserted patterns in discharge summaries
        :param line:
        """
        pattern1 = r'\*\*NAME\[.{0,20}\]'
        return re.sub(pattern1, '**NAME**', line)

    def prepare_text(self, text):
        """
        Read text in to list,  remove the empty lines
        Parameters:
            text - string of text to be split
        Return:
            result - list of strings from text
            :param text:
        """
        result = []
        for line in text.split('\n'):
            line = line.strip()
            if len(line) > 0:
                line = self.remove_double_star(line)
                line = self.clean(line)
                line = self.seperate_puncts(line)
                line = re.sub(r'[ ]+', ' ', line)
                line = line.strip()
                if len(line) > 0:
                    result.append(line)
        return result

    def seperate_puncts(self, line):
        """
        separate punctuations from words,  except '.'
        :param line:
        """
        linee = ''
        for w in line:
            if self.is_punct(w):
                if len(linee) > 0 and linee[-1] != ' ':
                    linee = linee + ' ' + w + ' '
                else:
                    linee = linee + w + ' '
            else:
                linee = linee + w
        return linee.strip()

    def is_digit(self, word):
        """
        Identify all numbers
        :param word:
        """
        if re.match(r'[+-]?\d*[.]?\d+$', word):  # all number
            return True
        return False

    def is_num_list(self, word):
        """
        for list like:  1.   2.
        :param word:
        """
        if re.match(r'\d+\.$', word):  # num list
            return True
        return False

    def is_punct(self, w):
        """
        indentify the punctuation
        :param w:
        """
        if w == ',' or w == '?' or w == '!' or w == '"' or w == ';' or w == ':':
            # '.' and '+' used in ABBR
            return True
        else:
            return False

    def clean(self, sentence):
        """
        Clean text: replace '[ ]+' into '[ ]'; remove '^[ ]+' and '[ ]+$'
        :param sentence:
        """
        sentence = re.sub(r'&amp;', '&', sentence)
        sentence = re.sub(r'&gt;|&gt[ ];|&lt;|&lt[ ];', ' ', sentence)
        sentence = re.sub(r'\'s|\'S', ' \'s', sentence)
        sentence = re.sub(r'[ ]+', ' ', sentence)
        return sentence

    def get_ave_len(self, text):
        """
        get the average length of all the lines
        :param text:
        """
        lenn = len(text)
        avel = 0
        for line in text:
            avel += len(line)
        return avel / lenn

    def is_stop_punct(self, w):
        """
        These puncts are definitely a sentence boundary
        :param w:
        """
        if w == '?' or w == '!' or w == '.' or w == ';':
            return True
        else:
            return False



import itertools
import os
from typing import Iterable

from pytakes import TextItem
from pytakes.iolib import sqlai


def get_next_from_directory(directories=None, encoding='utf8'):
    for directory in directories or ():
        for root, dirs, files in os.walk(directory):
            for file in files:
                fp = os.path.join(directory, file)
                try:
                    with open(fp, encoding=encoding) as fh:
                        text = fh.read()
                except FileNotFoundError:
                    continue
                else:
                    yield '.'.join(file.split('.')[:-1]) or file, text


def get_next_from_connections(*connections):
    for connection in connections:
        for doc_name, text in get_next_from_sql(**connection):
            yield doc_name, text


def get_next_from_sql(name=None, driver=None, server=None,
                      database=None, connection_string=None,
                      name_col=None, text_col=None):
    """
    :param name_col:
    :param text_col:
    :param name: tablename (if connecting to database)
    :param driver: db driver  (if connecting to database)
    :param server: name of server (if connecting to database)
    :param database: name of database (if connecting to database)
    """
    if name and driver and server and database:
        eng = sqlai.get_engine(connection_string=connection_string, driver=driver, server=server, database=database)
        for doc_name, text in eng.execute(f'select {name_col}, {text_col} from {name}'):
            yield doc_name, text


def get_next_from_corpus(*dirs, directories=None, encoding='utf8', connections=None) -> Iterable[TextItem]:
    for doc_name, text in itertools.chain(
            get_next_from_directory(dirs, encoding),
            get_next_from_directory(directories, encoding),
            get_next_from_connections(*connections or list())
    ):
        yield TextItem(text.split('\n\n'), file=doc_name)  # force new sentence with double-newline


import argparse

from pytakes.util import email_utils


def main():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('-s', '--subject', default='', help='Header for automated message.')
    parser.add_argument('-f', '--filename', default=None, help='Name of file to send.')
    parser.add_argument('-t', '--text', default=[], nargs='+', help='Text for email content.')
    parser.add_argument('-r', '--recipients', nargs='+', default=[], help='Recipients\' name,address.')
    parser.add_argument('--sender', default=None, help='Sender name,address.')
    parser.add_argument('--server-address', default=None, help='Mail server path.')
    args = parser.parse_args()
    email_utils.main(**vars(args))  # set up e-mail message and send


if __name__ == '__main__':  # if run from cmd line, not if imported
    main()



from pytakes.nlp.miner import Miner
from pytakes.nlp.terms import add_words


class MinerCollection(object):

    def __init__(self, ssplit=None, cleaner=None, add_intervening_words=True):
        self.miners = []
        self._ssplit = ssplit
        self.cleaner = cleaner
        self.add_words = add_intervening_words

    def add(self, miner: Miner):
        if miner is not None:
            self.miners.append(miner)

    def ssplit(self, text):
        if self._ssplit:
            # possible HACK? leaving original below, though perhaps
            #  something like split/join should be part of the sentence
            #  splitting algorithm itself
            # for sentence in self._ssplit(' '.join(text.split('\n'))):
            for sentence in self._ssplit(text):
                yield sentence
        else:
            yield text

    def clean_sentence(self, sentence):
        if self.clean_sentence and self.cleaner:
            return self.cleaner(sentence)
        else:
            return sentence

    def parse_sentence(self, sentence, offset):
        terms = []
        for miner in self.miners:
            terms += miner.clean(miner.mine(sentence, offset=offset))
        if self.add_words:
            terms += add_words(terms, sentence, offset=offset)
        terms.sort()  # otherwise, list is grouped by type
        for miner in self.miners:
            terms = miner.postprocess(terms)
        terms.sort()  # probably not required
        res = []
        for miner in self.miners:
            res += miner.extract(terms)
        return res, sentence

    def parse(self, doc):
        offset = 0
        for sentence in self.ssplit(doc.get_text()):
            sent = self.clean_sentence(sentence)
            yield self.parse_sentence(sent, offset)
            offset += len(sent)



import os

from pytakes.nlp.statusminer import StatusMiner


def test_load_from_csv():
    path = os.path.dirname(os.path.abspath(__file__))
    csv_file = os.path.join(path, 'data', 'negation.csv')
    rules = list(StatusMiner.load_negex_from_csv(csv_file))
    assert len(rules) == 12
    regex_loaded = False  # was regular expression correctly loaded
    for negex, type_, direction in rules:
        assert len(negex) > 0
        if r'\s' in negex:
            regex_loaded = True
        assert len(type_) == 4
        assert direction in {0, 1, 2, 3}
    assert regex_loaded



"""


Edit:
    2013-11-26    added direction, probability, hypothetical, etc. to Word and children
"""

from functools import total_ordering


@total_ordering
class Word(object):
    def __init__(self, word, begin, end, _type='term', offset=0):
        self.word_ = word
        self.begin_ = begin
        self.end_ = end
        self.type_ = _type.lower()
        self.offset_ = offset
        self.certainty_ = 4  # 0-4 (neg to affm)
        self.hypothetical_ = False  # for status only
        self.other_ = False  # for status only
        self.historical_ = False  # for status only
        self.direction_ = 0

    def get_absolute_begin(self):
        return self.begin() + self.offset_

    def get_absolute_end(self):
        return self.end() + self.offset_

    def is_negated(self, normal=True):
        if normal:
            return self.certainty_ == 0
        else:
            return not self.negated_

    def is_improbable(self):
        return self.certainty_ == 1

    def is_possible(self):
        return self.certainty_ == 2

    def is_probable(self):
        return self.certainty_ == 3

    def is_affirmed(self):
        return self.certainty_ == 4

    def get_certainty(self):
        return self.certainty_

    def set_certainty(self, level):
        self.certainty_ = level

    def is_hypothetical(self):
        return self.hypothetical_

    def is_historical(self):
        return self.historical_

    def is_not_patient(self):
        return self.other_

    def negate(self):
        self.certainty_ = 0

    def improbable(self):
        if self.certainty_ > 1:  # added 20140109
            self.certainty_ = 1

    def possible(self):
        if self.certainty_ > 2:  # added 20140109
            self.certainty_ = 2

    def probable(self):
        if self.certainty_ > 3:  # added 20140109
            self.certainty_ = 3

    def hypothetical(self):
        self.hypothetical_ = True

    def historical(self):
        self.historical_ = True

    def other_subject(self):
        self.other_ = True

    def direction(self):
        return self.direction_

    def begin(self):
        return self.begin_

    def set_begin(self, begin):
        self.begin_ = begin

    def end(self):
        return self.end_

    def set_end(self, end):
        self.end_ = end

    def word(self):
        return str(self.word_)

    def type(self):
        return self.type_

    def __len__(self):
        return self.end_ - self.begin_

    ''' Comparisons defined by relative location
        in the sentence. First term < last term. '''

    def __gt__(self, other):
        return self.begin_ > other.begin_ and self.end_ > other.end_  # must account for eq implementation

    def __eq__(self, other):
        """ equal if any overlap in indices """
        return (self.begin_ == other.begin_ or
                self.end_ == other.end_ or
                (self.begin_ > other.begin_ and self.end_ < other.end_) or
                (self.begin_ < other.begin_ and self.end_ > other.end_)
                )

    def __unicode__(self):
        return str(self.word_)

    def __str__(self):
        return str(self).encode('utf-8')

    def __repr__(self):
        return ('<' + str(self.word_) + ',' + str(self.begin_) + ':' +
                str(self.end_) +
                (',NEG' if self.is_negated() else ',POS' if self.is_possible else '') +
                ', <' + self.type_ + '>>')


class Term(Word):
    def __init__(self, word, begin, end, _type, id_, offset=0):
        super(Term, self).__init__(word, begin, end, _type, offset)
        self.id_ = id_

    def id(self):
        return self.id_

    def id_as_list(self):
        if isinstance(self.id_, list):
            return self.id_
        else:
            return [self.id_]

    def add_term(self, other):
        self.id_ = self.id_as_list() + other.id_as_list()


class Concept(Term):
    def __init__(self, word, begin, end, id_, cat, certainty=4,
                 hypothetical=0, historical=0, not_patient=0, offset=0):
        """
        For backwards compatibility:
            certainty used to be neg(ated), boolean
            hypothetical used to be pos(sible), boolean
        """
        super(Concept, self).__init__(word, begin, end, "concept", id_, offset=offset)
        self.cat_ = cat

        if isinstance(certainty, bool):  # old way
            neg = certainty
            pos = hypothetical
            if neg:
                self.negate()
            if pos:
                self.possible()

        else:  # new way
            self.set_certainty(certainty)
            if hypothetical:
                self.hypothetical()
            if historical:
                self.historical()
            if not_patient:
                self.other_subject()

    def cat(self):
        return self.cat_


class Negation(Word):
    def __init__(self, word, begin, end, _type='negn', direction=0, offset=0):
        super(Negation, self).__init__(word, begin, end, _type, offset=offset)
        self.direction_ = direction
        self.negate()


def find_terms(terms, text, offset=0):
    """
    Paramters:
        terms - list of (term,id) where unique id for each term
        :param terms:
        :param text:
        :param offset:
    """
    results = []
    for term, id_ in terms:
        for m in term.finditer(text):
            results.append(Term(term, m.start(), m.end(), 'term', id_, offset=offset))
    return sorted(results)


def clean_terms(terms):
    """
    remove terms that are subsumed by other terms
    :param terms:
    """
    terms.sort(reverse=False)
    if len(terms) < 2:
        return terms
    i = 1
    curr = 0
    while True:
        if terms[curr] == terms[i]:
            if len(terms[curr]) > len(terms[i]):
                del terms[i]
            elif len(terms[curr]) < len(terms[i]):
                del terms[curr]
                curr = i - 1
            elif terms[curr].begin() == terms[i].begin() and terms[curr].end() == terms[i].end():
                terms[curr].add_term(terms[i])
                del terms[i]
            else:  # keep both representations
                curr = i
                i += 1
        else:
            curr = i
            i += 1

        if i >= len(terms):
            return terms


def add_words(terms, text, offset=0):
    """
    Adds fill words between concepts (i.e., terms) from the text
    based on begin and end offsets.
    NB: Ignores extraneous words before and after concepts
    since these do not play any role in combining terms
    or determining negation.
    :param terms:
    :param text:
    :param offset:
    """
    curr = 0
    words = []
    for term in sorted(terms):
        if term.begin() > curr:
            for word in text[curr:term.begin()].split():
                words.append(Word(word, curr + 1, curr + 1 + len(word), offset=offset))
                curr = curr + 1 + len(word)
        curr = term.end()
    # ignoring extraneous words at end of sentence
    return words



import argparse
import csv
import logging
import logging.config
import os

from pytakes.dict.rules import read_file, read_rules, generate_combinations, build_rows
from pytakes.dict.rules import read_files
from pytakes.util import mylogger
from pytakes.util.db_reader import DbInterface
from pytakes.util.utils import get_valid_args


def check_if_table_exists(db, table_name):
    if db.is_sql_server_connection():
        return db.execute_fetchone('''
                    IF EXISTS (SELECT *
                               FROM INFORMATION_SCHEMA.TABLES
                               WHERE TABLE_NAME = '%s')
                      SELECT COUNT(*)
                        FROM %s
                    ELSE
                      SELECT -1
                    ''' % (table_name, table_name)
                                   )[0]
    else:  # unknown connection type: assume table doesn't exist
        logging.warning('Untested database connection: Connection is not SQL Server.')
        logging.warning('Assuming table "{}" does not exist. If it does exist, '
                        'delete it and restart the process.'.format(table_name))
        return -1


def drop_table(db, table_name):
    db.execute_commit('DROP TABLE %s;' % table_name)


def create_table(db, table_name):
    db.execute_commit('''
        CREATE TABLE %s
          (
            ID int,
            CUI varchar(15),
            Fword varchar(80),
            Text varchar(8000),
            Code varchar(45),
            SourceType varchar(45),
            TUI varchar(6),
            TextLength int,
            RegexVariation int,
            WordOrder int,
            Valence int
          )
    ''' % table_name)


def add_rows_to_db(rows, db, table_name):
    for num, (cui, fword, text, textlength, code, rxVar, wdOrder, val) in enumerate(rows):
        try:
            db.execute_commit('''
            INSERT INTO {} (ID, CUI, Fword, Text, TextLength, Code, SourceType, TUI, RegexVariation, WordOrder, Valence)
            VALUES ({}, '{}', '{}', '{}', {}, '{}', 'Custom', 'T033', {}, {}, {})
            '''.format(table_name, num, cui,
                       fword.replace("'", "''"), text.replace("'", "''"),
                       textlength, code, rxVar,
                       wdOrder, val))
        except Exception as e:
            return cui, fword, text, textlength, code, rxVar, wdOrder, val, str(e)
    return None


def add_rows_to_csv(rows, filename):
    with open(filename, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(
            ['ID', 'CUI', 'Fword', 'Text', 'TextLength', 'Code', 'RegexVariation', 'WordOrder', 'Valence', 'SourceType',
             'TUI'])
        for num, row in enumerate(rows):
            writer.writerow([num] + row + ['Custom', 'T033'])
    return True


class Generator(object):
    def __init__(self):
        self.cuis = set()
        self.cui_idx = 0
        self.codes = set()
        self.code_idx = 1

    def used_cuis(self, cuis):
        for cui in cuis:
            self.used_cui(cui)

    def used_cui(self, cui):
        self.cuis.add(cui)

    def used_codes(self, codes):
        for code in codes:
            self.used_code(code)

    def used_code(self, code):
        self.codes.add(code)

    def generate_cui(self):
        while True:
            new_cui = 'G%07d' % self.cui_idx
            self.cui_idx += 1
            if new_cui in self.cuis:
                continue
            else:
                return new_cui

    def generate_code(self):
        while True:
            new_code = '1%08d' % self.code_idx
            self.code_idx += 1
            if new_code in self.codes:
                continue
            else:
                return new_code


def create_sample_directory(path=None):
    if not path:
        path = ''
    d = os.path.join(path, 'cat')
    os.makedirs(d, exist_ok=True)
    with open(os.path.join(d, 'categories.txt'), 'w') as out:
        out.write('[good]\nSuperman\nBatman\n[evil]\n\nLoki\nTwo-Face||Two Face\n')
    with open(os.path.join(path, 'rules.txt'), 'w') as out:
        out.write('[good] vs [evil]\n[good] is a good guy\n[evil] is a villain\n')
    logging.info('Created sample directory structure.')


def build(path=None, output=None, table=None, driver=None, server=None, database=None,
          ignore_categories=None):
    cwd = path
    while not path:
        cwd = input('Parent directory of files: ')
        try:
            os.chdir(cwd)
        except Exception as e:
            logging.exception(e)
            logging.error('Failed to find directory "{}".'.format(cwd))
            path = None  # reset path so user gets prompt

    logging.info('Current directory: {}.'.format(cwd))
    generator = Generator()

    # try to find the rules and categories files
    files = os.listdir(cwd)

    categories = {}
    if ignore_categories:
        logging.info('Ignoring category files.')
    else:
        cat_dir = os.path.join(cwd, 'cat')
        if 'cat' in files:
            if os.path.isdir(cat_dir):
                logging.info('Found \'cat\' directory.')
                read_files([os.path.join(cat_dir, x) for x in os.listdir(cat_dir)], categories)
            else:
                read_file(cat_dir, categories)
        else:
            for file in files:
                file = os.path.join(cwd, file)
                if os.path.splitext(file)[1] == '.cat':
                    read_file(file, categories)
                    logging.info('Found category file: %s.' % file)
        logging.info('Found {} categories.'.format(len(categories)))

        if len(categories) == 0:
            ans = input('No categories were found, would you like\n' +
                        'an example created in your current directory? (Y/n) ')
            if ans.lower() in ['y', 'yes', 'ye']:
                create_sample_directory()
            logging.error('Could not find a category file. Terminating.')
            logging.info('If you are not using categories, consider using the "--ignore-categories" flag.')
            return

    rules = []
    for file in files:
        file = os.path.join(cwd, file)
        if os.path.splitext(os.path.split(file)[1])[0] == 'rules':
            rules += read_rules(file, generator)
            logging.info('Found rule file: %s.' % file)
    logging.info('Found %d rules.' % len(rules))

    if len(rules) == 0:
        logging.info('Please create a rules files (e.g., "rules.txt").')
        logging.info('Could not find any rules. Terminating.')
        return

    logging.info('Generating combinations.')
    combos = generate_combinations(rules, categories, generator)
    rows = build_rows(combos, generator)

    # option to create csv_file
    if output:
        csv_filename = output
    else:
        csv_filename = input('CSV filename (press enter to skip): ')

    if csv_filename:
        if csv_filename[-4:] != '.csv':
            csv_filename += '.csv'
        logging.info('Outputting rows to CSV: %s.' % csv_filename)
        add_rows_to_csv(rows, os.path.join(cwd, csv_filename))

    if table:
        newtable = table
    else:
        newtable = input('SQL Server table name (press enter to skip): ')
    if newtable:
        if not server and not database:
            logging.info('No database configuration specified for new table.')
            d = input('Specify driver for new table (press enter for default "{}")'.format(driver))
            if d:
                driver = d
        if not server:
            server = input('Specify server for new table: ')
        if not database:
            database = input('Specify database for new table: ')
        db = DbInterface(driver, server, database)
        while True:
            cnt = check_if_table_exists(db, newtable)
            if cnt >= 0:
                ans = input(('Table %s already exists with %d rows.\n' +
                             'Are you sure you want to drop and recreate\n' +
                             '%s? (Y/n) ') % (newtable, cnt, newtable))
                if ans.lower() not in ['y', 'yes', 'ye']:
                    # logic for 'no' answer
                    newtable = input('SQL Server table name (press enter to skip): ')
                    continue
                else:
                    logging.info('Dropping table %s.' % newtable)
                    try:
                        drop_table(db, newtable)
                    except Exception as e:
                        logging.exception('Exception while attempting to drop table.')
                        logging.exception(e)
                        raise e
            break
        try:
            create_table(db, newtable)
        except Exception as e:
            logging.exception('Exception while attempting to drop table.')
            logging.exception(e)
            raise e
        logging.info('Adding rows to table %s.' % newtable)
        res = add_rows_to_db(rows, db, newtable)
        if res:
            row_elements = [str(x) for x in res]
            exc_text = 'Exception while adding row: %s.' % ', '.join(row_elements)
            logging.exception(exc_text)
            raise Exception(exc_text)
    logging.info('Process completed successfully.')


def main():
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('-p', '--path', default=None, help='Parent directory of folders.')
    parser.add_argument('-o', '--output', default=None, help='Output csv file.')
    parser.add_argument('-t', '--table', default=None, help='Output database table.')
    parser.add_argument('--driver', default='SQL Server', help='Driver for database connection (if needed).')
    parser.add_argument('--server', default=None, help='Server for database connection (if needed).')
    parser.add_argument('--database', default=None, help='Database for database connection (if needed).')
    parser.add_argument('--create-sample', default=None, action='store_true', help='Create sample directory.')
    parser.add_argument('--ignore-categories', action='store_true', default=False,
                        help='Do not look for cat files (categories).')

    parser.add_argument('-v', '--verbosity', type=int, default=2, help='Verbosity of log output.')
    args = parser.parse_args()

    loglevel = mylogger.resolve_verbosity(args.verbosity)
    logging.config.dictConfig(mylogger.setup('builder', logdir=args.path, loglevel=loglevel))

    if args.create_sample:
        create_sample_directory(args.path)
    else:
        try:
            build(**get_valid_args(build, vars(args)))
        except Exception as e:
            logging.exception(e)


if __name__ == '__main__':
    main()



"""
Base classes for interface.
"""
import abc


class Dictionary(metaclass=abc.ABCMeta):
    def __init__(self, name=None, **kwargs):
        self.name = name

    @abc.abstractmethod
    def read(self):
        pass

    def int_or_default(self, val, default):
        try:
            return int(val)
        except (ValueError, TypeError):
            return default


class Document(metaclass=abc.ABCMeta):
    def __init__(self, name=None, **config):
        self.name = name

    @abc.abstractmethod
    def __len__(self):
        pass

    @abc.abstractmethod
    def get_ids(self):
        pass

    @abc.abstractmethod
    def read_next(self):
        pass


class Output(metaclass=abc.ABCMeta):
    # columns
    all_labels = ['dictid', 'concept', 'captured', 'context', 'certainty', 'hypothetical', 'historical',
                  'otherSubject', 'start_idx', 'end_idx', 'cpu_name', 'version']
    all_types = ['int', 'varchar(50)', 'varchar(255)', 'varchar(255)', 'int', 'int', 'int',
                 'int', 'int', 'int', 'varchar(50)', 'int']

    def __init__(self, name=None, context_width=75, **config):
        self.name = name
        self.context_width = context_width

    def __enter__(self):
        return self

    @abc.abstractmethod
    def writerow(self, feat, meta=None, text=None):
        pass

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @staticmethod
    def _get_index(length, value):
        if value < 0:
            return 0
        return min(value, length)

    @staticmethod
    def _get_text(text, start, end, remove_newlines=True):
        if remove_newlines:
            return ' '.join(text[start:end].split('\n'))
        else:
            return text[start:end]



from .csv import CsvDictionary, CsvDocument, CsvOutput
from .sas import SasDictionary, SasDocument
from .sql import SqlDictionary, SqlDocument, SqlOutput

data_items = {
    'sql': {
        'dictionary': SqlDictionary,
        'document': SqlDocument,
        'output': SqlOutput
    },
    'sas': {
        'dictionary': SasDictionary,
        'document': SasDocument,
        # 'output': SasOutput  # not supported
    },
    'csv': {
        'dictionary': CsvDictionary,
        'document': CsvDocument,
        'output': CsvOutput
    }
}


def get_data_item(res, datatype, conn):
    try:
        return data_items[res['type']][datatype](**res, dbi=conn)
    except Exception as e:
        raise ValueError('Unsupported {} type: "{}"'.format(res['type'], datatype))


import os

from pytakes.run.simple_run import run


def do_simple_run_and_get_outdir(outfile, negation_file=None, **kwargs):
    path = os.path.dirname(os.path.abspath(__file__))
    indir = os.path.join(path, 'data', 'files')
    outdir = os.path.join(path, 'data', 'testout')
    concepts = os.path.join(path, 'data', 'concepts.csv')
    negex_path = os.path.join(path, 'data', negation_file) if negation_file else None
    run(indir, outdir, concepts,
        outfile=outfile, negex_path=negex_path,
        hostname='Eumaeus', **kwargs)
    return outdir


def test_simple_run_csv():
    outfile = 'concepts.csv'
    outdir = do_simple_run_and_get_outdir(outfile=outfile)
    with open(os.path.join(outdir, outfile)) as fh:
        actual = fh.read()
    with open(os.path.join(outdir, 'expected.csv')) as fh:
        expected = fh.read()
    assert actual == expected


def test_simple_run_jsonl():
    outfile = 'concepts.jsonl'
    outdir = do_simple_run_and_get_outdir(outfile=outfile)
    with open(os.path.join(outdir, outfile)) as fh:
        actual = fh.read()
    with open(os.path.join(outdir, 'expected.jsonl')) as fh:
        expected = fh.read()
    assert actual == expected


def test_simple_run_jsonl_negex_csv():
    outfile = 'concepts.negex.jsonl'
    outdir = do_simple_run_and_get_outdir(
        outfile=outfile,
        negation_file='negation.csv'
    )
    with open(os.path.join(outdir, outfile)) as fh:
        actual = fh.read()
    with open(os.path.join(outdir, 'expected.negex.jsonl')) as fh:
        expected = fh.read()
    assert actual == expected



import json
import os
import platform

from .base import Output


class JsonlOutput(Output):

    def __init__(self, name=None, path=None, metalabels=None,
                 types=None, hostname=None, batch_number=None, **config):
        super().__init__(name=name, **config)
        self.labels = (metalabels or []) + self.all_labels
        self.types = types
        self.hostname = hostname or platform.node()
        self.batch_number = batch_number
        self.fp = os.path.join(path, self.name)
        self.fh = open(self.fp, 'w')

    def writerow(self, feat, meta=None, text=None):
        length = len(text)
        data = {
            'meta': meta,
            'concept_id': feat.id(),
            'concept': feat.cat(),
            'captured': self._get_text(text, feat.begin(), feat.end()),
            'context': self._get_text(text, self._get_index(length, feat.begin() - self.context_width),
                                      self._get_index(length, feat.end() + self.context_width)),
            'qualifiers': {
                'certainty': feat.get_certainty(),
                'hypothetical': feat.is_hypothetical(),
                'historical': feat.is_historical(),
                'other_subject': feat.is_not_patient(),
            },
            'start_index': feat.get_absolute_begin(),
            'end_index': feat.get_absolute_end(),
            'tracking': {
                'hostname': self.hostname,
                'batch': self.batch_number,
            }
        }
        self.fh.write(json.dumps(data) + '\n')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fh.close()



import logging
import os

try:
    from sas7bdat import SAS7BDAT
except ImportError:
    logging.warning(f'Missing SAS7BDAT library: will not be able to read from SAS.')
    SAS7BDAT = None

from .base import Dictionary, Document
from ..dict.textitem import TextItem


class SasDictionary(Dictionary):

    def __init__(self, path=None, valence=None,
                 regex_variation=None, word_order=None, **kwargs):
        super().__init__(**kwargs)
        self.path = path
        self.valence = valence
        self.regex_variation = regex_variation
        self.word_order = word_order
        self.fp = os.path.join(path, self.name)

    def read(self):
        logging.info('Getting Terms and Negation.')
        res = []
        with SAS7BDAT(self.fp) as fh:
            columns = fh.column_names_strings
            for row in fh:
                res.append([
                    row[columns['id']],
                    row[columns['text']],
                    row[columns['cui']],
                    row[columns['valence']],
                    row[columns['regexvariation']],
                    row[columns['wordorder']],
                ])
        return res


class SasDocument(Document):

    def __init__(self, path=None, order_by=None, batch_size=None, meta=None,
                 text=None, **config):
        super().__init__(**config)
        self.text = text
        self.meta = meta
        self.order_by = order_by
        self.batch_size = batch_size
        self.fp = os.path.join(path, self.name)

    def read_next(self):
        with SAS7BDAT(self.fp) as fh:
            for i, line in enumerate(fh):
                columns = {x: i for i, x in enumerate(line)}
                yield TextItem(
                    meta_list=[line[columns[m]] for m in self.meta],
                    text=[line[columns[t]] for t in self.text]
                )



"""
File: email_vmmc_log.py
Author: Scott Halgrim, halgrim.s@ghc.org
Date: 6/3/13
Functionality: Sends the latest info from the vmmc log so its status can
               more easily be checked.
Contents:
    LOGFN - log filename of delining process to send results from
    SERVER - smtp server address
    FROM - header info for sender
    TONAMES - list of names of sendees for header
    TOADDYS - list of e-mail addys for header and envelope
    TO - header info for sendees
    SUBJECT - subject of e-mail message
    getTodaysLines - function that get's the current day's lines out of a log
                     file into a list
    main - function that sets up the e-mail message with the appropriate message
           and sends
    __main__ code that calls main
TODO:
    - Make this more sophisticated
      - Store the last line sent and then use that to just get all the new lines
      - There may even be an SMTP handler in the logging stuff
NOTES:
    - Just copied from email_delining_log.py to create
"""
import logging
import smtplib
import email.utils
from email.mime.text import MIMEText
import argparse


def get_text_from_file(fn):
    """
    Function: getTodaysLines
    Input: fn - filename to get lines from
    Output: answer - a list of the lines (with training newlines) in fn where
                     a string representing today's date in the format yyyy-mm-dd
                     is present
    Functionality: Get's the current day's lines out of a log file into a list
    """
    with open(fn) as f:
        all_ = f.read()  # get all lines from file
    return all_  # return output


def resolve_recipients(recipients):
    resrecipients = []
    resaddr = []
    for rec in recipients:
        if ',' in rec:
            name, addr = rec.split(',')
        else:
            name, addr = rec, rec

        resrecipients.append(email.utils.formataddr((name, addr)))
        resaddr.append(addr)
    return resaddr, ';'.join(resrecipients)


def main(subject='', filename=None, text='',
         recipients=None, sender=None, server_address=None):
    """
    Functionality: Sets up e-mail message and sends

    :param subject:
    :param filename:
    :param recipients: list of string of "name,email@address"
    :param text:
    :param sender: string of "name,email@address"
    :param server_address: string
    """
    if not isinstance(text, str):
        text = '\n'.join(text)

    toaddys, to_ = resolve_recipients(recipients)

    _, from_ = resolve_recipients([sender])

    if filename:
        try:
            ftext = get_text_from_file(filename)
        except Exception as e:
            ftext = 'Failed to retrieve text from file "{}". {}'.format(filename, e)
    else:
        ftext = ''

        # create the text of the body
    msgstring = '''This is an automated email.
    
    %s

    %s\n'''

    # create MIME message from message body
    msg = MIMEText(msgstring % (text, ftext))
    msg['From'] = from_  # set header from info
    msg['To'] = to_  # set header to info
    msg['Subject'] = 'Automated Message: %s' % subject  # set header subject
    server = smtplib.SMTP(server_address)  # open connection to smtp server
    server.sendmail(toaddys[0], toaddys, msg.as_string())  # send the e-mail

    return


if __name__ == '__main__':  # if run from cmd line, not if imported
    parser = argparse.ArgumentParser(fromfile_prefix_chars='@')
    parser.add_argument('-s', '--subject', default='', help='Header for automated message.')
    parser.add_argument('-f', '--filename', default=None, help='Name of file to send.')
    parser.add_argument('-t', '--text', default=[], nargs='+', help='Text for email content.')
    parser.add_argument('-r', '--recipients', nargs='+', default=[], help='Recipients name,address.')
    parser.add_argument('--sender', default=None, help='Sender name,address.')
    parser.add_argument('--server-address', default=None, help='Mail server path.')
    args = parser.parse_args()
    main(**vars(args))  # set up e-mail message and send



import os
import tempfile

from pytakes.run.runner import run_config
from pytakes.run.schema import validate_config


def get_outpath(config):
    return os.path.join(config['output']['path'], config['output']['outfile'])


def test_run_config_jsonl():
    path = os.path.dirname(os.path.abspath(__file__))
    config_file = os.path.join(path, 'data', 'config', 'run.config.py')
    with tempfile.NamedTemporaryFile('w', suffix='.py', delete=False) as temp:
        with open(config_file) as fh:
            data = fh.read().replace('BASE_PATH', path)
        temp.write(data)
    config = validate_config(temp.name)
    run_config(temp.name)  # cannot run `run_config` directly, otherwise paths will not line up correctly
    with open(get_outpath(config)) as fh:
        actual = fh.read()
    with open(os.path.join(path, 'data', 'testout', 'expected.negex.jsonl')) as fh:
        expected = fh.read()
    assert actual == expected



import regex as re
from .utils import replace_punctuation, is_number

_1GRAM = '1gram'
_2GRAM = '2gram'
_3GRAM = '3gram'
_4GRAM = '4gram'
_S2GRAM = 's2gram'
_S3GRAM = 's3gram'
NUMBER = '<NUMBER>'
JOINER = '_'


class Feature(object):

    ID_COUNTER = 0

    def __init__(self, feature, category):
        self.id = self._get_id()
        self.feature = feature
        self.category = category

    def _get_id(self):
        self.ID_COUNTER += 1
        return self.ID_COUNTER

    def get_feature(self):
        return self.feature

    def get_category(self):
        return self.category

    def get_id(self):
        return self.id


class Excluder:
    def __init__(self, stopwords=None, patterns=None):
        """
        Create an instance used to exclude features from consideration.
        :param stopwords: skip these words when collecting features
        :param patterns: do not include any features with this pattern
        :return:
        """
        self.stopwords = set(stopwords)
        self.patterns = [re.compile(p) for p in patterns]
        if self.stopwords or self.patterns:
            self.excluding = True
        else:
            self.excluding = False

    def __contains__(self, item):
        return item in self.stopwords

    def exclude(self, feature, joiner):
        if self.patterns:
            for pat in self.patterns:
                for feat in feature.split(joiner) + [feature]:
                    if pat.match(feat):
                        return True
        return False


class FeatureMiner(object):

    def __init__(self, stopwords=None, patterns=None, number_norm=True):
        self.excluder = Excluder(stopwords, patterns)
        self.number_norm = number_norm

    def _add_feature(self, features, ngram, category, joiner=JOINER):
        if not self.excluder.exclude(ngram, joiner):
            features.append(Feature(ngram, category))

    def mine(self, sentences):
        """
        computes all types of ngrams in one pass

        features: dict of feature -> count
        sentences: list of sentences: list of strings
        number_norm: whether to normalize numbers or not
        excluder: instance of feature.Excluder
        """
        all_features = []
        for sentence in sentences:
            pppw = None
            ppw = None
            pw = None
            us = JOINER
            features = []
            for word in replace_punctuation(sentence).split():
                word = word.lower()
                if self.excluder.excluding and word in self.excluder:
                    continue  # ignore word
                elif self.number_norm and is_number(word):
                    word = NUMBER
                self._add_feature(features, word, _1GRAM)  # unigram
                if pw:
                    self._add_feature(features, us.join((pw, word)), _2GRAM)  # bigram
                    if ppw:
                        self._add_feature(features, us.join((ppw, pw, word)), _3GRAM)  # trigram
                        self._add_feature(features, us.join((ppw, word)), _S2GRAM)  # split bigram
                        if pppw:
                            self._add_feature(features, us.join((pppw, word)), _S2GRAM)  # split bigram
                            self._add_feature(features, us.join((pppw, ppw, word)), _S3GRAM)  # split trigram
                            self._add_feature(features, us.join((pppw, pw, word)), _S3GRAM)  # split trigram
                            self._add_feature(features, us.join((pppw, ppw, pw, word)), _4GRAM)  # 4gram
                # assignments for next pass
                pppw = ppw
                ppw = pw
                pw = word
            all_features.append(features)
        return all_features



from distutils.core import setup
import setuptools

setup(name='pytakes',
      version='1.0.5',
      description='Basic information extraction tool.',
      url='https://bitbucket.org/dcronkite/pytakes',
      author='dcronkite',
      license='MIT',
      classifiers=[  # from https://pypi.python.org/pypi?%3Aaction=list_classifiers
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: Text Processing :: Linguistic',
          'License :: OSI Approved :: MIT License',
      ],
      keywords='nlp information extraction',
      entry_points={
          'console_scripts':
              [
                  'pytakes-automate-run = pytakes.automate_run:main',
                  'pytakes-negex-creator = pytakes.negex_creator:main',
                  'pytakes-postprocessor = pytakes.postprocessor:main',
                  'pytakes-processor = pytakes.processor:main',
                  'pytakes-sendmail = pytakes.sendmail:main',
                  'pytakes-build-dictionary = pytakes.build_dictionary:main',
              ]
      },
      install_requires=['regex'],
      package_dir={'': 'src'},
      packages=setuptools.find_packages('src'),
      package_data={'pytakes.nlp': ['data/*.db']},
      zip_safe=False
      )
